# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the classes and methods from pyretis.core.box"""
import logging
import unittest
import numpy as np
from pyretis.core.box import (
    create_box,
    RectangularBox,
    TriclinicBox,
    box_matrix_to_list,
    box_vector_angles,
    angles_from_box_matrix,
    box_from_restart
)
logging.disable(logging.CRITICAL)


class RectBoxTest(unittest.TestCase):
    """Run the tests for the RectangularBox and the generic box."""

    def test_create_empty_box(self):
        """Test the creation of boxes with no arguments."""
        box = create_box()
        self.assertIsInstance(box, RectangularBox)
        self.assertEqual(box.low, [-float('inf')])
        self.assertEqual(box.high, [float('inf')])
        self.assertEqual(box.length, [float('inf')])
        self.assertEqual(box.periodic, [False])

    def test_create_missing_periodic(self):
        """Test default behaviour of creation without periodic arguments."""
        box = create_box(cell=[10, 10, 10], periodic=None)
        self.assertIsInstance(box, RectangularBox)
        self.assertEqual(box.periodic, [True, True, True])
        self.assertTrue(np.allclose(box.low, [0., 0., 0.]))
        self.assertTrue(np.allclose(box.high, [10., 10., 10.]))
        self.assertTrue(np.allclose(box.length, [10., 10., 10.]))

        box = create_box(cell=[10, 10, 10], periodic=[False])
        self.assertEqual(box.periodic, [False, True, True])

    def test_create_box_fail(self):
        """Test some corner cases when we fail to create a box."""
        # Test that we fail for too many periodic values given:
        with self.assertRaises(ValueError):
            create_box(cell=[10, 10, 10], periodic=[False]*4)
        # Test that we fail when low == high:
        with self.assertRaises(ValueError):
            create_box(cell=[10, 10], low=[0, 10], high=[10, 10])
        # Test that we fail for inconsistent lengths:
        with self.assertRaises(ValueError):
            create_box(cell=[10, 10], low=[0, 0], high=[10, 20])

    def test_box_size(self):
        """Test that giving a size works as expected."""
        test_in = (
            {'length': [10, 10]},
            {'low': [0., -10.], 'high': [10., 20.]},
            {'low': [1, 1, 1], 'length': [10, 11, 12]},
            {'high': [1, 1, 1], 'length': [10, 11, 12]},
            {'high': [10, 9, 8]},
            {'low': [-10, 10]},
        )
        correct = (
            {'low': [0., 0.], 'high': [10., 10.], 'length': [10., 10.]},
            {'low': [0., -10.], 'high': [10., 20.], 'length': [10., 30.]},
            {'low': [1, 1, 1], 'high': [11., 12., 13.],
             'length': [10., 11., 12.]},
            {'low': [-9, -10, -11], 'high': [1., 1., 1.],
             'length': [10., 11., 12.]},
            {'low': [0., 0., 0.], 'high': [10, 9, 8], 'length': [10, 9, 8]},
            {'low': [-10, 10], 'high': [float('inf'), float('inf')],
             'length': [float('inf'), float('inf')]},
        )
        for case, corr in zip(test_in, correct):
            box = create_box(low=case.get('low', None),
                             high=case.get('high', None),
                             cell=case.get('length', None),
                             periodic=case.get('periodic', None))
            self.assertIsInstance(box, RectangularBox)
            self.assertTrue(np.allclose(box.length, corr['length']))
            self.assertTrue(np.allclose(box.low, corr['low']))
            self.assertTrue(np.allclose(box.high, corr['high']))

    def test_volume_calculate(self):
        """Test calculation of volume."""
        box = create_box(cell=[10])
        self.assertIsInstance(box, RectangularBox)
        self.assertAlmostEqual(box.calculate_volume(), 10.)
        box2 = create_box(cell=[10, 10])
        self.assertIsInstance(box2, RectangularBox)
        self.assertAlmostEqual(box2.calculate_volume(), 100.)
        box3 = create_box(cell=[10, 10, 10])
        self.assertIsInstance(box3, RectangularBox)
        self.assertAlmostEqual(box3.calculate_volume(), 1000.)
        box4 = create_box(low=[-10, 0, 11], high=[10, 5, 19])
        self.assertIsInstance(box4, RectangularBox)
        self.assertAlmostEqual(box4.calculate_volume(), 20.*5.*8.)

    def test_faulty_input(self):
        """Test that the initialisation fails as we expect."""
        with self.assertRaises(ValueError):
            create_box(cell=[10, -10, 10])
        with self.assertRaises(TypeError):
            create_box(cell=10)
        with self.assertRaises(ValueError):
            create_box(cell=[10, (-10, 10), (0, -15)])
        with self.assertRaises(ValueError):
            create_box(low=[0, 0], high=[10, 10], cell=[11, 11])
        with self.assertRaises(ValueError):
            create_box(cell=['crash', 15])
        with self.assertRaises(ValueError):
            create_box(low=[10, 10], high=[10, 10])

    def test_update_box(self):
        """Test update of box size."""
        box = create_box(cell=[10, 10, 10])
        new_length = [10, 11, 12]
        box.update_size(new_length)
        for i, j in zip(box.length, new_length):
            self.assertAlmostEqual(i, j)
        new_length2 = [13, 12, 11, 10]
        box.update_size(new_length2)  # This should NOT update.
        for i, j in zip(box.length, new_length):
            self.assertAlmostEqual(i, j)
        new_length3 = [3, 3]
        box.update_size(new_length3)  # This should NOT update.
        for i, j in zip(box.length, new_length):
            self.assertAlmostEqual(i, j)
        new_length4 = None
        box.update_size(new_length4)
        for i, j in zip(box.length, new_length):
            self.assertAlmostEqual(i, j)

    def test_bounds(self):
        """Test the bounds method."""
        box = create_box(cell=[10, 11, 12])
        correct = [[0., 10.], [0., 11.], [0., 12.]]
        bounds = box.bounds()
        for bound, corr in zip(bounds, correct):
            for i, j in zip(bound, corr):
                self.assertAlmostEqual(i, j)

    def test_print_length(self):
        """Test that we print out cell parameters correctly."""
        lengths = [
            [1],
            [1, 2],
            [1, 2, 3],
            [1, 2, 3, 4, 5, 6],
            [1, 2, 3, 4, 5, 6, 7, 8, 9],
        ]
        fmt = '{:6.2f}'
        for i in lengths:
            box = create_box(cell=i)
            correct = ' '.join([fmt.format(j) for j in i])
            self.assertEqual(correct, box.print_length(fmt=fmt))

    def test_restart(self):
        """Test that we create and read restart info for a box."""
        box_settings = [
            {'low': [10, -1, 101], 'high': [12, 5, 102],
             'periodic': [True, True, False], 'cell': [2., 6., 1.]},
            {'low': [9, 8, 7], 'periodic': [False, True, False],
             'cell': [1, 2, 3, 4, 5, 6]},
        ]

        keys = ('cell', 'periodic', 'low', 'high')

        for setting in box_settings:
            box = create_box(**setting)
            restart = box.restart_info()
            self.assertTrue(all(k in restart for k in keys))
            for key, val in restart.items():
                if key not in setting:
                    continue
                if key == 'periodic':
                    self.assertEqual(val, setting[key])
                else:
                    self.assertTrue(np.allclose(val, setting[key]))
            info = {'box': restart}
            box2 = box_from_restart(info)
            # todo The line under or above should be removed once the restart
            # strategy is unified.
            box2.load_restart_info(restart)
            self.assertTrue(box == box2)
            self.assertFalse(box != box2)

    def test_pbc_coordinate_dim(self):
        """Test pbc for a specific dimension."""
        length = [10, 11, 12]
        box = create_box(cell=length, periodic=[False, True, True])
        pos = [11, 10, 14]
        pbc_pos = box.pbc_coordinate_dim(pos[2], 2)
        self.assertAlmostEqual(pbc_pos, 2.0)
        pbc_pos = box.pbc_coordinate_dim(pos[0], 0)
        self.assertAlmostEqual(pbc_pos, pos[0])

    def test_pbc_wrap(self):
        """Test pbc wrap for coordinates."""
        length = [10, 11, 12]
        box = create_box(cell=length, periodic=[False, True, True])
        pos = np.array([[11, 10, 14], ])
        correct = np.array([[11, 10, 2], ])
        pbc_pos = box.pbc_wrap(pos)
        self.assertTrue(np.allclose(correct, pbc_pos))
        self.assertFalse(np.allclose(correct, pos))

    def test_pbc_dist_matrix(self):
        """Test pbc wrap for a distance vector."""
        box = create_box(cell=[10, 10, 10], periodic=[False, True, True])
        dist = np.array([[8., 7., 9.], ])
        pbc_dist = box.pbc_dist_matrix(dist)
        correct = np.array([[8., -3., -1.], ])
        self.assertTrue(np.allclose(correct, pbc_dist))
        self.assertTrue(np.allclose(correct, dist))


class TriBoxTest(unittest.TestCase):
    """Run the tests specific for the TriclinicBox class."""

    def test_create_box(self):
        """Test creation of TriclinicBox."""
        box1 = create_box(
            cell=[17.5092040633036, 7.58170825120892, 6.95903807579504,
                  4.37730063346742, 0.0, 0.0]
        )
        self.assertIsInstance(box1, TriclinicBox)
        box2 = create_box(cell=[10, 10, 10, 0.0, 0.0, 0.0])
        self.assertIsInstance(box2, TriclinicBox)
        box3 = create_box(cell=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertIsInstance(box3, TriclinicBox)

    def test_volume_calculate(self):
        """Test calculation of volume."""
        box1 = create_box(cell=[10, 1, 1, 0, 0, 0])
        self.assertAlmostEqual(box1.calculate_volume(), 10.)
        box2 = create_box(cell=[10, 11, 12, 0, 0, 0])
        self.assertAlmostEqual(box2.calculate_volume(), 10.*11.*12.)
        length = [17.5092040633036, 7.58170825120892, 6.95903807579504,
                  4.37730063346742, 0.0, 0.0]
        box3 = create_box(cell=length)
        self.assertAlmostEqual(box3.calculate_volume(), 923.810056228)
        length = [17.5092040633036, 7.58170825120892, 6.95903807579504,
                  4.37730063346742, 0.0, 0.0, 0.0, 0.0, 0.0]
        box4 = create_box(cell=length)
        self.assertAlmostEqual(box4.calculate_volume(), 923.810056228)

    def test_update_box(self):
        """Test update for triclinic box."""
        box = create_box(cell=[1, 2, 3, 0, 0, 0])
        new_size = [1, 2, 3, 4, 5, 6]
        box.update_size(new_size)
        correct_size = np.array([[1., 4., 5.], [0., 2., 6.], [0., 0, 3.]])
        self.assertTrue(np.allclose(box.box_matrix, correct_size))
        new_size = [-1, -1, -1]
        box.update_size(new_size)  # This should NOT update size.
        self.assertTrue(np.allclose(box.box_matrix, correct_size))
        new_size = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        box.update_size(new_size)
        correct_size = np.array([[1., 4., 5.], [6., 2., 7.], [8., 9., 3.]])
        self.assertTrue(np.allclose(box.box_matrix, correct_size))

    def test_box_matrix_to_list(self):
        """Test that we return the box matrix as a list as expected."""
        new_size = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        box = create_box(cell=new_size)
        out = box_matrix_to_list(box.box_matrix)
        for i, j in zip(new_size, out):
            self.assertAlmostEqual(i, j)
        self.assertIsNone(box_matrix_to_list(None))

    def test_box_vector_angles(self):
        """Test conversion from a,b,c, alpha, beta, gamma."""
        test_data = [
            {'length': [3., 3., 3.],
             'alpha': 100., 'beta': 80., 'gamma': 75.,
             'box': np.array([[3.0, 0.77646, 0.52094],
                              [0.0, 2.89778, -0.67891],
                              [0.0, 0.0, 2.87536]])},
            {'length': [3., 5.1, 1.9],
             'alpha': 100., 'beta': 85., 'gamma': 66.,
             'box': np.array([[3.0, 2.07436, 0.16559],
                              [0.0, 4.65908, -0.43488],
                              [0.0, 0.0, 1.84213]])},
        ]
        for i in test_data:
            box_matrix = box_vector_angles(i['length'], i['alpha'],
                                           i['beta'], i['gamma'])
            self.assertTrue(np.allclose(box_matrix, i['box'], atol=1e-4))

    def test_get_angles(self):
        """Test that we can get box angles and lengths from a box matrix."""
        test_data = [
            {'length': [3., 3., 3.],
             'alpha': 100., 'beta': 80., 'gamma': 75.,
             'box': np.array([[3.0, 0.77646, 0.52094],
                              [0.0, 2.89778, -0.67891],
                              [0.0, 0.0, 2.87536]])},
            {'length': [3., 5.1, 1.9],
             'alpha': 100., 'beta': 85., 'gamma': 66.,
             'box': np.array([[3.0, 2.07436, 0.16559],
                              [0.0, 4.65908, -0.43488],
                              [0.0, 0.0, 1.84213]])},
        ]
        for i in test_data:
            length, alpha, beta, gamma = angles_from_box_matrix(i['box'])
            self.assertTrue(np.allclose(length, i['length'], atol=1e-5))
            self.assertAlmostEqual(alpha, i['alpha'], places=3)
            self.assertAlmostEqual(beta, i['beta'], places=3)
            self.assertAlmostEqual(gamma, i['gamma'], places=3)


class BoxTest(unittest.TestCase):
    """Run the test for the base class."""

    def test_box_equality(self):
        """Test that we can compare boxes."""
        box1 = create_box(cell=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        box2 = create_box(cell=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(box1, box2)
        # Test failure for different periodic settings:
        box1.periodic = [False, True, True]
        self.assertNotEqual(box1, box2)
        box1.periodic = [False]
        self.assertNotEqual(box1, box2)
        box1.periodic = [True, True, True]
        self.assertEqual(box1, box2)
        box1.length = None
        self.assertNotEqual(box1, box2)
        box2.length = None
        self.assertEqual(box1, box2)
        del box2.length
        self.assertNotEqual(box1, box2)
        box2.length2 = 100
        self.assertNotEqual(box1, box2)
        box3 = create_box(cell=[1, 2])
        self.assertNotEqual(box1, box3)
        box4 = create_box(cell=[1, 2])
        self.assertEqual(box3, box4)
        box3.dim = 100
        self.assertNotEqual(box3, box4)

    def test_box_copy(self):
        """Test that we can copy a box."""
        box1 = create_box(cell=[1, 2, 3, 4, 5, 6, 7, 8, 9])
        box2 = box1.copy()
        self.assertIsNot(box1, box2)
        self.assertEqual(box1, box2)
        box1 = create_box(cell=[1, 2, 3])
        box2 = box1.copy()
        self.assertIsNot(box1, box2)
        self.assertEqual(box1, box2)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Extension to color the name of the app + commands in code snippets."""
from pygments.lexer import inherit
from pygments.lexers.shell import BashLexer
from pygments.token import Name
from sphinx import highlighting


KEYWORDS = (
    'pip',
    'pip3',
    'git',
    'f2py',
    'pyretisrun',
    'pyretisanalyse',
    'virtualenv',
    'conda',
    'sudo',
    'mkdir',
    'pylint',
    'pycodestyle',
    'pydocstyle',
)


class PyretisLexer(BashLexer):
    """Add a few keyword tokes. These will be colored."""

    name = 'Pyretis in bash'
    aliases = ['pyretis']
    mimetypes = ['application/x-sh', 'application/x-shellscript']
    merged = '|'.join(KEYWORDS)
    tokens = {
        'basic': [
            (r'\b({})(?=[\s)`])'.format(merged),
             Name.Builtin), inherit,
        ],
    }


def setup(app):
    pass


highlighting.lexers['pyretis'] = PyretisLexer()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module defines common methods for the settings handling.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

create_external (:py:func:`.create_external`)
    Method to create objects from settings.

check_settings (:py:func:`.check_settings`)
    Check that required simulation settings are actually given.

create_engine (:py:func:`.create_engine`)
    Method to create an engine from settings.

create_orderparameter (:py:func:`.create_orderparameter`)
    Method to create order parameters from settings.

create_potential (:py:func:`.create_potential`)
    Method to create a potential from settings.
"""
import logging
import os
from pyretis.core.common import initiate_instance, import_from
from pyretis.engines import engine_factory
from pyretis.orderparameter import order_factory
from pyretis.orderparameter.orderparameter import CompositeOrderParameter
from pyretis.forcefield.factory import potential_factory
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['create_external', 'check_settings', 'create_orderparameter',
           'create_engine', 'create_potential']


def check_settings(settings, required):
    """Check that required simulation settings are actually given.

    This method will look for required settings in the given
    `settings`. If one or more keys from the given `required` list of
    strings are not found, this method will return False. Otherwise,
    it will return True. Typically, an exception should be raised if
    False is returned, this is handled outside the method in case
    someone wants to add some magic handling of missing settings.

    Parameters
    ----------
    settings : dict
        This dict contains the given settings
    required : list of strings
        This list contains the settings that are required and which
        we will check the presence of.

    Returns
    -------
    result : boolean
        True if all required settings are present, False otherwise.
    not_found : list of strings
        There are the required settings we did not find.

    """
    result = True
    not_found = []
    for setting in required:
        if setting not in settings:
            result = False
            not_found.append(setting)
    return result, not_found


def create_external(settings, key, factory, required_methods,
                    key_settings=None):
    """Create external objects from settings.

    This method will handle the creation of objects from settings. The
    requested objects can be PyRETIS internals or defined in external
    modules.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.
    key : string
        The setting we are creating for.
    factory : callable
        A method to call that can handle the creation of internal
        objects for us.
    required_methods : list of strings
        The methods we need to have if creating an object from external
        files.
    key_settings : dict, optional
        This dictionary contains the settings for the specific key we
        are processing. If this is not given, we will try to obtain
        these settings by `settings[key]`. The reason why we make it
        possible to pass these as settings is in case we are processing
        a key which does not give a simple setting, but a list of settings.
        It that case `settings[key]` will give a list to process. That list
        is iterated somewhere else and `key_settings` can then be used to
        process these elements.

    Returns
    -------
    out : object
        This object represents the class we are requesting here.

    """
    if key_settings is None:
        try:
            key_settings = settings[key]
        except KeyError:
            logger.debug('No "%s" setting found. Skipping set-up', key)
            return None
    module = key_settings.get('module', None)
    klass = None
    try:
        klass = key_settings['class']
    except KeyError:
        logger.debug('No "class" setting for "%s" specified. Skipping set-up',
                     key)
        return None
    if module is None:
        return factory(key_settings)
    # Here we assume we are to load from a file. Before we import
    # we need to check that the path is ok or if we should include
    # the 'exe_path' from settings.
    # 1) Check if we can find the module:
    if os.path.isfile(module):
        obj = import_from(module, klass)
    else:
        if 'exe_path' in settings['simulation']:
            module = os.path.join(settings['simulation']['exe_path'],
                                  module)
            obj = import_from(module, klass)
        else:
            msg = f'Could not find module "{module}" for {key}!'
            raise ValueError(msg)
    # run some checks:
    for function in required_methods:
        objfunc = getattr(obj, function, None)
        if not objfunc:
            msg = f'Could not find method {klass}.{function}'
            logger.critical(msg)
            raise ValueError(msg)
        if not callable(objfunc):
            msg = f'Method {klass}.{function} is not callable!'
            logger.critical(msg)
            raise ValueError(msg)
    return initiate_instance(obj, key_settings)


def create_orderparameter(settings):
    """Create order parameters from settings.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.

    Returns
    -------
    out : object like :py:class:`.OrderParameter`
        This object represents the order parameter.

    """
    main_order = create_external(
        settings,
        'orderparameter',
        order_factory,
        ['calculate'],
    )
    if main_order is None:
        logger.info('No order parameter created')
        return None
    logger.info('Created main order parameter:\n%s', main_order)

    extra_cv = []
    order_settings = settings.get('collective-variable', [])
    for order_setting in order_settings:
        order = create_external(
            settings,
            'collective-variable',
            order_factory,
            ['calculate'],
            key_settings=order_setting
        )
        logger.info('Created additional collective variable:\n%s', order)
        extra_cv.append(order)
    if not extra_cv:
        return main_order
    all_order = [main_order] + extra_cv
    order = CompositeOrderParameter(order_parameters=all_order)
    logger.info('Composite order parameter:\n%s', order)
    return order


def create_engine(settings):
    """Create an engine from settings.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.

    Returns
    -------
    out : object like :py:class:`.EngineBase`
        This object represents the engine.

    """
    engine = create_external(settings, 'engine', engine_factory,
                             ['integration_step'])
    if not engine:
        raise ValueError('Could not create engine from settings!')
    return engine


def create_potential(settings, key_settings):
    """Create a potential from settings.

    Parameters
    ----------
    settings : dict
        This dictionary contains the settings for the simulation.
    key_settings : dict
        Settings for the potential we are creating.

    Returns
    -------
    out : object like :py:class:`.PotentialFunction`
        The object representing the potential function.

    """
    return create_external(settings, 'potential', potential_factory,
                           ['force', 'potential', 'potential_and_force'],
                           key_settings=key_settings)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A script for building the C Velocity Verlet extension."""
from setuptools import (
    setup,
    Extension,
)
import numpy as np


CMODULE = Extension(
    'vvintegrator',
    sources=['vvintegrator.c'],
    extra_compile_args=['-Ofast', '-march=native']
)
setup(
    name='PyRETIS Velocity Verlet C extension',
    description='C extension for the Velocity Verlet integrator.',
    ext_modules=[CMODULE],
    include_dirs=[np.get_include()],
)



from openmm import app
import openmm as mm
from openmm import unit

pdb = app.PDBFile('input.pdb')
forcefield = app.ForceField('amber99sbildn.xml', 'tip3p.xml')

system = forcefield.createSystem(pdb.topology, nonbondedMethod=app.PME,
                                 nonbondedCutoff=1.0*unit.nanometers,
                                 constraints=app.HBonds, rigidWater=True,
                                 ewaldErrorTolerance=0.0005)
integrator = mm.LangevinIntegrator(300*unit.kelvin,
                                   1.0/unit.picoseconds,
                                   2.0*unit.femtoseconds)
integrator.setConstraintTolerance(0.00001)

# Check with "python -m simtk.testInstallation" which platforms are available
# Here I use CUDA, but CPU is more commonly available
# platform = mm.Platform.getPlatformByName('CPU')
platform = mm.Platform.getPlatformByName('CUDA')
simulation = app.Simulation(pdb.topology, system, integrator, platform)
simulation.context.setPositions(pdb.positions)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of interaction with a Simulation."""
from pyretis.core import System
from pyretis.simulation import Simulation


def function(sys):
    """Return the set temperature for the system."""
    return sys.temperature


simulation = Simulation(settings={}, controls={'steps': 10})
system = System(temperature=300)

my_task = {'func': function,
           'args': [system],
           'first': True,
           'result': 'temperature'}

simulation.add_task(my_task)

for result in simulation.run():
    step = result['cycle']['step']
    temp = result['temperature']['set']
    print(step, temp)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Methods that might be useful for several tests.

This module defines methods that can be used for comparing results.

"""
import math
import os
import filecmp
import numpy as np
from pyretis.testing.helpers import search_for_files
from pyretis.inout.formats.energy import EnergyPathFile
from pyretis.inout.formats.order import OrderPathFile
from pyretis.inout.formats.path import PathExtFile


# Names of the expected output files in archive directories:
ARCHIVE_FILES = {'energy.txt', 'order.txt', 'traj.txt'}
# Names of other expected output files:
OUTPUT_FILES = {'energy.txt', 'order.txt', 'pathensemble.txt'}
# Define readers for loading data:
READERS = {
    'energy': EnergyPathFile,
    'order': OrderPathFile,
    'traj': PathExtFile,
}


def read_files(*files, read_comments=True):
    """Read files into memory.

    Here, we assume that we are given small files and that we
    can read these into memory.

    Parameters
    ----------
    files : list of strings
        These are the paths to the files we are to read.
    read_comments : boolean
        If False, we skip lines starting with a "#".

    Returns
    -------
    out : list of list of strings
        The data read from the different files.

    """
    all_data = []
    for filename in files:
        data = []
        with open(filename, 'r', encoding="utf8") as infile:
            for line in infile:
                if not read_comments and line.strip().startswith('#'):
                    continue
                data.append(line)
        all_data.append(data)
    return all_data


def compare_files_lines(file1, file2, skip=None):
    """Compare two files, line by line.

    Parameters
    ----------
    file1 : string
        The path to the first file to compare.
    file2 : string
        The path to the second file to compare.
    skip : list of integers, optional
        These are line numbers we are to skip.

    Returns
    -------
    out[0] : boolean
        True if the files are deemed to be equal.
    out[1] : string
        A descriptive message of the result of the comparison.

    """
    all_data = read_files(file1, file2, read_comments=True)
    assert len(all_data) == 2
    data1, data2 = all_data[0], all_data[1]
    if len(data1) != len(data2):
        return False, 'The number of lines in the files differ'
    for i, (lini, linj) in enumerate(zip(data1, data2)):
        if skip and i in skip:
            continue
        if not lini == linj:
            return False, f'Line {i} differs: {lini.strip()} != {linj.strip()}'

    return True, 'Files are equal'


def compare_files_columns(file1, file2, file_type, skip=None):
    """Compare two output PyRETIS files.

    This method compares files where numbers are stored in columns
    and the columns have specific labels. Here, we also compare
    labels and comments.

    Parameters
    ----------
    file1 : string
        The path to the first file to compare.
    file2 : string
        The path to the second file to compare.
    file_type : string
        A string used to determine the file type.
    skip : list of strings, optional
        A list of items from the loaded data we are to skip.
        This can, for instance, be certain energy terms that are
        not absolute and can't easily be compared.

    Returns
    -------
    out[0] : boolean
        True if the files are deemed to be equal.
    out[1] : string
        A descriptive message of the result of the comparison.

    """
    reader = READERS[file_type]
    data1 = reader(file1, 'r').load()
    data2 = reader(file2, 'r').load()
    # Compare the files by compare the block found in the file:
    for block1, block2 in zip(data1, data2):
        # Start with block comments:
        if block1['comment'] != block2['comment']:
            return False, 'Block comment differs'
        # Compare terms found in the blocks:
        if sorted(block1['data'].keys()) != sorted(block2['data'].keys()):
            return False, 'Different items in block data'
        # Compare numerical data:
        for key, val in block1['data'].items():
            if skip and key in skip:
                continue
            if not np.allclose(val, block2['data'][key]):
                return False, 'Block terms differ'
    return True, 'Files are equal'


def compare_files_numerical(file1, file2):
    """Compare two output PyRETIS files.

    Here, we compare files that contain numerical data. We don't
    care about comments here, we just compare the actual numerical data.

    Parameters
    ----------
    file1 : string
        The path to the first file to compare.
    file2 : string
        The path to the second file to compare.

    Returns
    -------
    out[0] : boolean
        True if the files are deemed to be equal.
    out[1] : string
        A descriptive message of the result of the comparison.

    """
    data1 = np.loadtxt(file1)
    data2 = np.loadtxt(file2)
    if not np.allclose(data1, data2):
        return False, 'Numerical data differ'
    return True, 'Files are equal'


def compare_files(file1, file2, skip=None, mode='line'):
    """Compare two files.

    Parameters
    ----------
    file1 : string
        The path to the first file to compare.
    file2 : string
        The path to the second file to compare.
    skip : list of strings or list of ints, optional
        A list of items that are to be skipped in the comparison.
    mode : string
        A string used to determine how we do the comparison:
        ``'numerical'`` will select a comparison in which the
        file is parsed and numerical data compared;
        ``'line'`` will select a line-by-line comparison;
        anything else will perform a comparison using
        :py:func:`filecmp.cmp`.

    Returns
    -------
    out[0] : boolean
        True if the files were found to be equal, False otherwise.
    out[1] : string
        A string with information about the comparison result.

    """
    if mode == 'numerical':
        equal, msg = compare_files_numerical(file1, file2)
    elif mode == 'line':
        equal, msg = compare_files_lines(file1, file2, skip=skip)
    else:
        equal = filecmp.cmp(file1, file2, shallow=False)
        msg = 'Files are equal' if equal else 'Files are not equal'
    return equal, msg


def compare_traj_archive(dir1, dir2):
    """Compare archived trajectories.

    These archives consist of trajectory information such as
    energies, order parameters and positions. Here, we will not
    compare the actual raw trajectory data, but we verify that
    the output written by PyRETIS is identical in the two cases.

    Parameters
    ----------
    dir1 : string
        The path to the first directory to use in the comparison.
    dir2 : string
        The path to the second directory to use in the comparison.

    Returns
    -------
    out : list of tuples
        This list contains the files which differed, if any.

    """
    errors = []
    files1 = sorted(search_for_files(dir1))
    files2 = sorted(search_for_files(dir2))
    # Are the number of files equal:
    if len(files1) != len(files2):
        errors.append((dir1, dir2))
        return errors
    # Compare the files that are written by PyRETIS:
    for file1, file2 in zip(files1, files2):
        basename1 = os.path.basename(file1)
        basename2 = os.path.basename(file2)
        if basename1 != basename2:
            errors.append((file1, file2))
            continue
        if basename1 in ARCHIVE_FILES:
            equal, _ = compare_files(file1, file2, mode='cmp')
            if not equal:
                errors.append((file1, file2))
    return errors


def compare_pathensemble_files(file1, file2, rel_tol=1e-5, skip=None):
    """Compare two path ensemble files.

    We compare line-by-line, but skip comments and we check that
    numbers are close, as judged by the given relative tolarance.

    Parameters
    ----------
    file1 : string
        The path to the first file to consider in the comparison.
    file2 : string
        The path to the second file to consider in the comparison.
    rel_tol : float, optional
        A relative tolerance which is used to determine if numbers
        are almost equal.
    skip : list of integers, optional
        These are columns we are to skip in the comparison.

    Returns
    -------
    out[0] : boolean
        True if the files are equal, False otherwise.
    out[1] : string
        A message describing the result of the comparison.

    """
    all_data = read_files(file1, file2, read_comments=False)
    assert len(all_data) == 2
    if not len(all_data[0]) == len(all_data[1]):
        return False, 'The number of lines in the files differ'
    # Define the expected data types for the columns in the path
    # ensemble files:
    data_types = {
        0: int, 1: int, 2: int, 3: str, 4: str, 5: str, 6: int, 7: str, 8: str,
        9: float, 10: float, 11: int, 12: int, 13: float, 14: int, 15: int,
    }
    for i, (line1, line2) in enumerate(zip(*all_data)):
        stuff1 = line1.split()
        stuff2 = line2.split()
        for col, func in data_types.items():
            if skip and col in skip:
                continue
            if func == str:
                check = func(stuff1[col]) == func(stuff2[col])
            else:
                check = math.isclose(
                    func(stuff1[col]), func(stuff2[col]), rel_tol=rel_tol
                )
            if not check:
                return False, f'Files differ on line {i}, column {col}'

    return True, 'Files are equal'



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Some common methods for the tests."""
from contextlib import contextmanager
from io import StringIO
import logging
import numpy as np
from unittest.mock import patch
from pyretis.core.box import create_box
from pyretis.core.particles import Particles
from pyretis.core.random_gen import MockRandomGeneratorBorg
from pyretis.core.system import System
from pyretis.engines.internal import MDEngine
from pyretis.forcefield import ForceField, PotentialFunction
from pyretis.setup.createsimulation import create_ensembles
from pyretis.inout.settings import (fill_up_tis_and_retis_settings,
                                    add_default_settings,
                                    add_specific_default_settings)
from pyretis.simulation.path_simulation import (
    SimulationRETIS,
    SimulationTIS,
)


@contextmanager
def turn_on_logging():
    """Turn on logging so that tests can detect it."""
    logging.disable(logging.NOTSET)
    try:
        yield
    finally:
        logging.disable(logging.CRITICAL)


class MockEngine(MDEngine):
    """Create a fake engine for testing."""

    def __init__(self, timestep):
        """Set up the engine."""
        super().__init__(timestep, 'MockEngine', dynamics='Fake')
        self.direction = 1.0
        self.steps = 0
        self.reverse_after = 10

    def integration_step(self, system):
        """Do a fake integration step."""
        system.particles.pos += self.timestep * self.direction
        self.steps += 1
        if self.steps > self.reverse_after:
            self.steps = 0
            self.reverse_after += 10
            self.direction *= -1.0
        system.potential_and_force()

    def reset(self):
        """Reset attributes to initial values."""
        self.direction = 1.0
        self.steps = 0
        self.reverse_after = 10


class MockEngineOneWay(MDEngine):
    """Create a fake engine for testing."""

    def __init__(self, timestep):
        """Set up the engine."""
        super().__init__(timestep, 'MockEngineForward', dynamics='Fake')

    def integration_step(self, system):
        """Do a fake integration step."""
        system.particles.pos += self.timestep
        system.potential_and_force()


class MockEngineVelocitySupremacist(MDEngine):
    """Create a fake engine for testing."""

    def __init__(self, timestep):
        """Set up the engine."""
        super().__init__(timestep, 'MockEngineVelocitySupremacist',
                         dynamics='Fake')

    def integration_step(self, system):
        """Do a fake integration step."""
        if (system.particles.vel < 0).all():
            pass
        else:
            system.particles.pos += self.timestep
        system.potential_and_force()


class MockPotential(PotentialFunction):
    """Create a fake potential for testing."""

    def __init__(self):
        super().__init__(dim=1, desc='A fake potential')

    def potential(self, system):
        """Return the position of the particles."""
        return system.particles.pos.sum()

    def force(self, system):
        """Return the fake force and virial."""
        forces = system.particles.pos * -1.0
        virial = np.zeros((self.dim, self.dim))
        return forces, virial

    def potential_and_force(self, system):
        """Return the fake potential, force and virial."""
        pot = self.potential(system)
        forces, virial = self.force(system)
        return pot, forces, virial


def create_system():
    """Set up a system for testing."""
    box = create_box(periodic=[False])
    system = System(units='reduced', temperature=1.0, box=box)
    system.particles = Particles(dim=1)
    system.add_particle(-1.0 * np.ones((1, 1)))
    system.forcefield = ForceField(
        'Mock force field', potential=[MockPotential()]
    )
    return system


def create_test_retis_simulation():
    """Create a simple RETIS simulation."""
    system = create_system()
    engine = MockEngine(0.0123)

    settings = {
        'simulation': {
            'task': 'retis',
            'rgen': 'mock-borg',
            'interfaces': [-0.9, -0.5, 0.0]},
        'engine': {'obj': engine},
        'system': {'obj': system},
        'orderparameter': {'class': 'PositionVelocity',
                           'index': 0,
                           'dim': 'x',
                           'periodic': False},
        'tis': {
            'freq': 0.5,
            'maxlength': 20000,
            'aimless': True,
            'allowmaxlength': False,
            'sigma_v': -1,
            'seed': 1,
            'rgen': 'mock-borg',
            'zero_momentum': False,
            'rescale_energy': False,
        },
        'retis': {
            'rgen': 'mock-borg',
            'nullmoves': True,
        }
    }
    add_default_settings(settings)
    add_specific_default_settings(settings)
    fill_up_tis_and_retis_settings(settings)

    with patch('sys.stdout', new=StringIO()):
        ensembles = create_ensembles(settings)
        simulation = SimulationRETIS(ensembles, settings,
                                     {'rgen': 'mock-borg'})
    return simulation


def create_test_tis_simulation(engine_type='MockEngine', maxlength=20000):
    """Create a simple TIS simulation."""
    system = create_system()
    engines = {
        'MockEngine': MockEngine,
        'MockEngineOneWay': MockEngineOneWay,
        'MockEngineVelocitySupremacist': MockEngineVelocitySupremacist,
    }
    engine = engines.get(engine_type)(0.0123)
    settings = {
        'simulation': {
            'task': 'tis',
            'interfaces': [-0.9, -0.5, 0.0]},
        'engine': {'obj': engine},
        'system': {'obj': system},
        'orderparameter': {'class': 'PositionVelocity',
                           'index': 0,
                           'dim': 'x',
                           'periodic': False},
        'tis': {
            'freq': 0.5,
            'maxlength': maxlength,
            'aimless': True,
            'allowmaxlength': False,
            'sigma_v': -1,
            'seed': 1,
            'rgen': 'to-be-ignored-and-replaced-below',
            'zero_momentum': False,
            'rescale_energy': False,
        },
    }
    add_default_settings(settings)
    add_specific_default_settings(settings)
    fill_up_tis_and_retis_settings(settings)

    with patch('sys.stdout', new=StringIO()):
        ensembles = create_ensembles(settings)
    rgen_class = MockRandomGeneratorBorg.make_new_swarm()
    for ens in ensembles:
        ens['rgen'] = rgen_class(seed=0)
        ens['system'].order = ens['order_function'].calculate(ens['system'])

    simulation = SimulationTIS(
        ensembles,
        settings,
        {'rgen': rgen_class(seed=0)}
    )
    return simulation



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Defining a collapse block for the PyRETIS web-page."""
from docutils import nodes
from pathlib import Path
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged
from jinja2 import Template


HTML_PRE = """
<p>
<a class="btn btn-primary btn-sm" data-toggle="collapse" href="#{{ id }}">
{{ heading }} &raquo;</a>
  <div id="{{ id }}" class="collapse">"""
TEMPLATE_PRE = Template(HTML_PRE)
HTML_POST = """  </div></p>"""
TEMPLATE_POST = Template(HTML_POST)


class PyretisCollapseNode(nodes.General, nodes.Element):
    """Create a new node for the collapse block contents."""
    pass


class PyretisCollapseBlock(Directive):
    """Define a new directive."""

    required_arguments = 0
    optional_arguments = 0
    option_spec = {'heading': unchanged}
    has_content = True

    def run(self):
        """Set up a new node for a collapse block."""
        env = self.state.document.settings.env
        rst_source = Path(self.state_machine.node.document['source'])
        rst_filename = rst_source.name
        target_id = "%s.pyretiscb-%d" % (
            rst_filename,
            env.new_serialno('pyretis-ccb')
        )
        target_id = target_id.replace(".", "-")
        # target_node = nodes.target('', '', ids=[target_id])

        node = PyretisCollapseNode()
        node['heading'] = self.options.get('heading', 'title')
        node['target_id'] = target_id
        self.state.nested_parse(self.content, self.content_offset, node)
        return [node]


def html_visit_block(self, node):
    """Render the template when the block is visited."""
    self.body.append(
        TEMPLATE_PRE.render(id=node['target_id'],
                            heading=node['heading'])
    )


def html_depart_block(self, _):
    """End the HTML code when the block is done."""
    self.body.append(TEMPLATE_POST.render())


def setup(app):
    """Register the collable block."""
    app.add_node(
        PyretisCollapseNode,
        html=(
            html_visit_block,
            html_depart_block
        )
    )
    app.add_directive('pyretis-collapse-block', PyretisCollapseBlock)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the FORTRAN implementation of the Velocity Verlet integrator.

This test is comparing:

1) The Python (numpy) implementation

2) The FORTRAN implementation.
"""
# pylint: disable=invalid-name
import unittest
import numpy as np
from pyretis.core import System, create_box, Particles
from pyretis.simulation import Simulation
from pyretis.core.units import create_conversion_factors
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials import PairLennardJonesCutnp
from pyretis.tools import generate_lattice
from pyretis.engines import VelocityVerlet
from vvintegratorf import VelocityVerletF


def create_positions():
    """Create particles for the test."""
    create_conversion_factors('lj')
    lattice, size = generate_lattice('fcc', [3, 3, 3], density=0.9)
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1],
                     periodic=[True, True, True])
    npart = len(lattice)
    lattice += np.random.randn(npart, 3) * 0.05
    particles = Particles(dim=3)
    for pos in lattice:
        particles.add_particle(pos, np.zeros_like(pos), np.zeros_like(pos),
                               mass=1.0, name='Ar', ptype=0)
    msg = 'Created lattice with {} atoms.'
    print(msg.format(particles.npart))
    return particles, box


def run_test(steps, integrator, system=None):
    """Execute a test MD simulation."""
    if system is None:
        # create system
        particles, box = create_positions()
        system = System(temperature=1.0, units='lj', box=box)
        system.particles = particles
        initial = system.copy()
    else:
        initial = None
    parameters = {0: {'sigma': 1.0, 'epsilon': 1.0, 'rcut': 2.5}}
    potentialnp = PairLennardJonesCutnp(dim=3, shift=True)
    forcefieldnp = ForceField('Lennard-Jones force field',
                              potential=[potentialnp], params=[parameters])
    system.forcefield = forcefieldnp
    controls = {'steps': steps}
    simulation = Simulation(settings=None, controls=controls)
    task_integrate = {'func': integrator.integration_step,
                      'args': [system]}
    simulation.add_task(task_integrate)
    traj = []
    for _ in simulation.run():
        traj.append(
            {
                'pos': system.particles.get_pos(),
                'vel': system.particles.get_vel(),
                'force': system.particles.get_force()
            }
        )
    return traj, initial, system


class VVIntegratorTest(unittest.TestCase):
    """Run the tests for the FORTRAN integrator."""

    def test_integrator(self):
        """Test by integrating the equations of motion."""
        integrator = VelocityVerlet(0.0025)
        traj, initial, system = run_test(20, integrator)
        # reset to initial state
        system = initial
        # repeat with external integrator:
        integratorf = VelocityVerletF(0.0025)
        traj2, _, _ = run_test(20, integratorf, system)
        for trj1, trj2 in zip(traj, traj2):
            posok = np.allclose(trj1['pos'], trj2['pos'])
            self.assertTrue(posok)
            velok = np.allclose(trj1['vel'], trj2['vel'])
            self.assertTrue(velok)
            forceok = np.allclose(trj1['force'], trj2['force'])
            self.assertTrue(forceok)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package contains functions for setting up plotters.

Specifically, it defines colors, color schemes and a function
for selecting a plotter.

Here we also hard-code some color schemes which may be useful for
the plotting:

* the colorblind 10 scheme [cb10]_,

* the deep color scheme from the seaborn project [deep]_ and

* the husl color scheme [husl]_.

Package structure
-----------------

Modules
~~~~~~~

__init__.py
    This file. Handles imports for PyRETIS and defines some colors
    and the method for creating a plotter.

mpl_plotting.py (:py:mod:`pyretis.inout.plotting.mpl_plotting`)
    Methods for generating plots using matplotlib.

plotting.py (:py:mod:`pyretis.inout.plotting.plotting`)
    Definition of a generic base class for the plotter(s).

txt_plotting.py (:py:mod:`pyretis.inout.plotting.txt_plotting`)
    Methods for generating text output.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

create_plotter (:py:func:`.create_plotter`)
    Method to create a plotter.

Folders
~~~~~~~

styles
    This folder contains style files for matplotlib which are used by
    PyRETIS.

References
~~~~~~~~~~

.. [cb10] The colorblind_10 color scheme,
          https://jiffyclub.github.io/palettable/tableau/
.. [deep] The deep color scheme, from the seaborn project
          http://stanford.edu/~mwaskom/software/seaborn/index.html
.. [husl] The husl color scheme, http://www.husl-colors.org/

"""
from pyretis.inout.settings import SECTIONS
from .mpl_plotting import MplPlotter, mpl_set_style
from .txt_plotting import TxtPlotter


__all__ = ['create_plotter']


# Custom named colors:
COLORS = {'almost_black': '#262626'}
# Custom color-schemes. The default will be defined by the style file.
# The husl schemes are suited when many different colors are needed. They
# are hard-coded with the different number of colors.
COLOR_SCHEME = {'colorblind_10': ['#006BA4', '#FF800E', '#ABABAB', '#595959',
                                  '#5F9ED1', '#C85200', '#898989', '#A2C8EC',
                                  '#FFBC79', '#CFCFCF'],
                'deep': ['#4C72B0', '#55A868', '#C44E52', '#8172B2',
                         '#CCB974', '#64B5CD'],
                'husl_10': ['#f67088', '#db8831', '#ad9c31', '#77aa31',
                            '#33b07a', '#35aca4', '#38a8c5', '#6e9af4',
                            '#cc79f4', '#f565cc'],
                'husl_15': ['#f67088', '#f37932', '#ca9131', '#ad9c31',
                            '#8ea531', '#4fb031', '#33b07a', '#34ad99',
                            '#36abae', '#38a8c5', '#3ba3ec', '#9491f4',
                            '#cc79f4', '#f45fe3', '#f569b7'],
                'husl_20': ['#f67088', '#f7754f', '#db8831', '#c29431',
                            '#ad9c31', '#96a331', '#77aa31', '#31b23e',
                            '#33b07a', '#34ae92', '#35aca4', '#36abb3',
                            '#38a8c5', '#3aa5de', '#6e9af4', '#a38cf4',
                            '#cc79f4', '#f45bf1', '#f565cc', '#f66bad']}


def create_plotter(plot_settings, out_dir=None):
    """Create a plotter from given settings.

    The input plot settings is assumed to be a dictionary which we use
    for creating the plotter. In case the plot settings are not given, we
    just return None. We are here assuming that they are not given
    simply because we do not want to create a plotter.

    Parameters
    ----------
    plot_settings : dict
        These are the settings to create the plotter from. Here, we
        look for the keys `plotter`, `output` and `style` which defines
        the plotter to use, the output format and the style to use.
    out_dir : string, optional
        This string selects if and where the output should be written
        to. It's mainly used for internal purposes to save the report
        files to a specific directory.

    Returns
    -------
    out : object like :py:class:`.MplPlotter`
        This is an object which can be used for plotting.

    """
    default = SECTIONS['analysis']['plot']
    plotter = plot_settings.get('plotter', default['plotter'])
    out_fmt = plot_settings.get('output', default['output'])
    style = plot_settings.get('style', default['style'])
    backup = plot_settings.get('backup', SECTIONS['output']['backup'])
    if plotter.lower() in ['mpl', 'matplotlib']:
        return MplPlotter(out_fmt, backup=backup, style=style,
                          out_dir=out_dir)
    msg = f'Unknown plotter: {plotter}'
    raise ValueError(msg)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""pyvisa - An application for analysing PyRETIS simulations.

This script is a part of the PyRETIS library and can be used for
analysing the result from simulations.

usage: pyvisa.py [-h]

optional arguments:
  -cmp --pyvisa_compressor  use only the compressor tool without GUI.
  -data --pyvisa-data       select the data source.
  -h, --help                show this help message and exit.
  -i INPUT, --input INPUT   location of PyRETIS input files
                            or PyVisA compressed file.
  -oo --only-order          read only the order.txt files.
  -V, --version             show program's version number and exit.

"""
# pylint: disable=invalid-name
import argparse
import datetime
import logging
import os
import sys
import colorama  # pylint: disable=import-error
from pyretis import __version__ as VERSION
from pyretis.bin.pyretisanalyse import write_traceback
from pyretis.pyvisa.info import PROGRAM_NAME, CITE, LOGO, URL
from pyretis.inout.common import (
    check_python_version)
from pyretis.inout.formats.formatter import (
    LOG_FMT,
    PyretisLogFormatter,
)
from pyretis.inout import print_to_screen
from pyretis.pyvisa.orderparam_density import PathDensity, pyvisa_compress
from pyretis.pyvisa.common import recalculate_all
from pyretis.__init__ import HAS_PYVISA_REQ
if HAS_PYVISA_REQ:
    from pyretis.pyvisa.visualize import visualize_main  # pragma: no cover


# Set up for logging:
logger = logging.getLogger('')
logger.setLevel(logging.DEBUG)

runpath = os.getcwd()
ERROR_FILE = 'pyvisa_error.txt'


def hello_pyvisa(run_dir, infile):
    """Output a standard greeting for PyVISA.

    Parameters
    ----------
    run_dir : string
        The location where we are executing the analysis.
    infile : string
        String showing the location of the input file.

    """
    pyversion = sys.version.split()[0]
    msgtxt = [LOGO]
    msgtxt += ['                                                    Starting']
    msgtxt += [f'{PROGRAM_NAME} version: {VERSION}']
    msgtxt += [f'Python version: {pyversion}']
    msgtxt += [f'Running in directory: {run_dir}']
    msgtxt += [f'Input file: {infile}']
    print_to_screen('\n'.join(msgtxt), level='message')
    logger.info('\n'.join(msgtxt))


def bye_pyvisa():
    """Print out the goodbye message for PyVisA."""
    _DATE_FMT = '%d.%m.%Y %H:%M:%S'
    timeend = datetime.datetime.now().strftime(_DATE_FMT)
    msgtxt = f'End of {PROGRAM_NAME} execution: {timeend}'
    print_to_screen(msgtxt, level='info')
    logger.info(msgtxt)
    # display some references:
    references = [f'{PROGRAM_NAME} references:']
    references.append(('-')*len(references[0]))
    for line in CITE.split('\n'):
        if line:
            references.append(line)
    reftxt = '\n'.join(references)
    logger.info(reftxt)
    print_to_screen()
    print_to_screen(reftxt)
    urltxt = str(URL)
    logger.info(urltxt)
    print_to_screen()
    print_to_screen(urltxt, level='info')


def pyvisa_visual(basepath, input_file, pyvisa_dict):  # pragma: no cover
    """Load data to PyVisA.

    Parameters
    ----------
    basepath : string
        The execution folder where the input files are.
    input_file : string
        The input file with settings for the analysis.
    pyvisa_dict : dictionary, optional
        It determines the section of pyvisa to use, it contains:

    """
    # pylint: disable=possibly-used-before-assignment
    if not HAS_PYVISA_REQ:
        msg = ('Requirements not installed. You can still generate the '
               'hdf5 by using the -cmp flag instead')
        raise ImportError(msg)

    if input_file is None:
        visualize_main(basepath)
    # If a compressed file is given as an input
    elif input_file.endswith(('.hdf5', '.zip')):
        visualize_main(basepath, input_file)

    elif input_file.endswith('rst'):
        # If -data is selected with a trajectory
        if pyvisa_dict['pyvisa_data'] is not None:
            visualize_main(basepath, input_file,
                           trajfile=pyvisa_dict['pyvisa_data'])
        else:
            p_data = PathDensity(basepath, iofile=input_file)
            p_data.infos['only_ops'] = pyvisa_dict['only_order']

            # Walk selected ensemble directories
            p_data.walk_dirs(only_ops=pyvisa_dict['only_order'])
            visualize_main(basepath, p_data, rst_file=input_file)

    else:
        msg = f'The format of {input_file} is not supported.'
        raise ImportError(msg)


def main(basepath, input_file, pyvisa_dict=None):
    """Run the analysis.

    Parameters
    ----------
    basepath : string
        The execution folder where the input files are.
    input_file : string
        The input file with settings for the analysis.
    pyvisa_dict : dictionary, optional
        It determines the section of pyvisa to use, it contains:

         * `pyvisa_cmp`, boolean
           If true, only the compressor tool will be executed. A compressed
           file will be produced.
         * `pyvisa_visual`, boolean
           If true, only the visualization tool will be executed.
         * `pyvisa_data`, str
           If given, the file or folder containing the files that
           will be used to feed to PyVisA.
         * `pyvisa_recalculate`, boolean
           If true, use the recalculation tool to compute new op
           and cv values.

    """
    try:
        if input_file is not None and \
                not os.path.isfile(os.path.join(basepath, input_file)):
            raise FileNotFoundError(f'Input file "{input_file}" NOT found!')
        # For recalculation of op and cv's
        if pyvisa_dict.get('pyvisa_recalculate', False):
            recalculate_all(basepath, input_file,
                            data=pyvisa_dict['pyvisa_data'])
        # For compression of the files ONLY
        elif pyvisa_dict.get('pyvisa_compressor', False):
            pyvisa_compress(basepath, input_file, pyvisa_dict)
        else:
            pyvisa_visual(basepath, input_file, pyvisa_dict)

    # pylint: disable=broad-exception-caught
    except Exception as error:  # pragma: no cover
        errtxt = f'{type(error).__name__}: {error.args}'
        print_to_screen(errtxt, level='error')
        print_to_screen('Execution failed!', level='error')
        print_to_screen(f'Error traceback is written to: {ERROR_FILE}',
                        level='error')
        write_traceback(os.path.join(basepath, ERROR_FILE))
    finally:
        bye_pyvisa()


def entry_point():  # pragma: no cover
    """entry_point - The entry point for the pip install of pyretisanalyse."""
    colorama.init(autoreset=True)
    parser = argparse.ArgumentParser(description=PROGRAM_NAME)
    parser.add_argument('-i', '--input',
                        help=(f'Location of {PROGRAM_NAME} input file'),
                        required=False, default=None)
    parser.add_argument('-V', '--version', action='version',
                        version=f'{PROGRAM_NAME} {VERSION}')
    parser.add_argument('-cmp', '--pyvisa_compressor',
                        action='store_true',
                        help='Run PyVisA compressor tool',
                        default=False)
    parser.add_argument('-data', '--pyvisa-data',
                        action='store',
                        help='Select PyVisA data',
                        default=None)
    parser.add_argument('-recalculate', '--pyvisa-recalculate',
                        action='store_true',
                        help='Recalculate op and cv data.',
                        default=False)
    parser.add_argument('-oo', '--only_order', action='store_true',
                        help=('PyVisA: Use only data from order.txt files'),
                        default=False)

    args_dict = vars(parser.parse_args())
    # Define a console logger. This will log to sys.stderr:
    console = logging.StreamHandler()
    console.setLevel(logging.WARNING)
    console.setFormatter(PyretisLogFormatter(LOG_FMT))
    logger.addHandler(console)

    check_python_version()
    infile = args_dict['input']
    basepath = os.getcwd() if infile is None else os.path.dirname(
        os.path.abspath(infile))
    hello_pyvisa(basepath, infile)
    main(basepath, infile, args_dict)


if __name__ == '__main__':  # pragma: no cover
    entry_point()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A collection of potential functions.

This package defines some potential functions. These potential functions
can be used to create force fields.


Package structure
-----------------

Modules
~~~~~~~

potentials.py (:py:mod:`pyretis.forcefield.potentials.potentials`)
    This module defines some simple potential functions.

Sub-packages
~~~~~~~~~~~~

pairpotentials (:py:mod:`pyretis.forcefield.potentials.pairpotentials`)
    This package defines different pair interactions, for instance the
    Lennard-Jones 6-12 simple cut potential.

Important classes defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

DoubleWell (:py:class:`.DoubleWell`)
    A double well potential

RectangularWell (:py:class:`.RectangularWell`)
    A rectangular well potential -- useful as a bias potential.

PairLennardJonesCut (:py:class:`.PairLennardJonesCut`)
    The Lennard-Jones potential in pure python.

PairLennardJonesCutnp (:py:class:`.PairLennardJonesCutnp`)
    The Lennard-Jones potential, making use of numpy.

DoubleWellWCA (:py:class:`.DoubleWellWCA`)
    A n-dimensional Double Well potential.
"""
from .potentials import DoubleWell, RectangularWell
from .pairpotentials import (PairLennardJonesCut, PairLennardJonesCutnp,
                             DoubleWellWCA)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A CP2K external MD integrator interface.

This module defines a class for using CP2K as an external engine.

Important classes defined here
------------------------------

CP2KEngine (:py:class:`.CP2KEngine`)
    A class responsible for interfacing CP2K.
"""
import logging
import os
import re
import shlex
from pyretis.engines.external import ExternalMDEngine
from pyretis.core.random_gen import create_random_generator
from pyretis.inout.settings import look_for_input_files
from pyretis.core.particlefunctions import (
   calculate_kinetic_energy,
   reset_momentum
)
from pyretis.inout.formats.xyz import (
    read_xyz_file,
    write_xyz_trajectory,
    convert_snapshot
)
from pyretis.inout.formats.cp2k import (
    update_cp2k_input,
    read_cp2k_restart,
    read_cp2k_box,
    read_cp2k_energy,
)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


OUTPUT_FILES = {
    'energy': '{}-1.ener',
    'restart': '{}-1.restart',
    'pos': '{}-pos-1.xyz',
    'vel': '{}-vel-1.xyz',
    'wfn': '{}-RESTART.wfn',
    'wfn-bak': '{}-RESTART.wfn.bak-'
}


REGEXP_BACKUP = re.compile(r'\.bak-\d$')


def write_for_step_vel(infile, outfile, timestep, subcycles, posfile, vel,
                       name='md_step', print_freq=None):
    """Create input file for a single step.

    Note, the single step actually consists of a number of subcycles.
    But from PyRETIS' point of view, this is a single step.
    Further, we here assume that we start from a given xyz file and
    we also explicitly give the velocities here.

    Parameters
    ----------
    infile : string
        The input template to use.
    outfile : string
        The file to create.
    timestep : float
        The time-step to use for the simulation.
    subcycles : integer
        The number of sub-cycles to perform.
    posfile : string
        The (base)name for the input file to read positions from.
    vel : numpy.array
        The velocities to set in the input.
    name : string, optional
        A name for the CP2K project.
    print_freq : integer, optional
        How often we should print to the trajectory file.

    """
    if print_freq is None:
        print_freq = subcycles
    to_update = {
        'GLOBAL': {
            'data': [f'PROJECT {name}',
                     'RUN_TYPE MD',
                     'PRINT_LEVEL LOW'],
            'replace': True,
        },
        'MOTION->MD':  {
            'data': {'STEPS': subcycles,
                     'TIMESTEP': timestep}
        },
        'MOTION->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
        'MOTION->PRINT->RESTART->EACH': {
            'data': {'MD': print_freq}
        },
        'MOTION->PRINT->VELOCITIES->EACH': {
            'data': {'MD': print_freq}
        },
        'MOTION->PRINT->TRAJECTORY->EACH': {
            'data': {'MD': print_freq}
        },
        'FORCE_EVAL->SUBSYS->TOPOLOGY': {
            'data': {'COORD_FILE_NAME': posfile,
                     'COORD_FILE_FORMAT': 'xyz'}
        },
        'FORCE_EVAL->SUBSYS->VELOCITY': {
            'data': [],
            'replace': True,
        },
        'FORCE_EVAL->DFT->SCF->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
    }
    for veli in vel:
        to_update['FORCE_EVAL->SUBSYS->VELOCITY']['data'].append(
            f'{veli[0]} {veli[1]} {veli[2]}'
        )
    remove = [
        'EXT_RESTART',
        'FORCE_EVAL->SUBSYS->COORD'
    ]
    update_cp2k_input(infile, outfile, update=to_update, remove=remove)


def write_for_integrate(infile, outfile, timestep, subcycles, posfile,
                        name='md_step', print_freq=None):
    """Create input file for a single step for the integrate method.

    Here, we do minimal changes and just set the time step and subcycles
    and the starting configuration.

    Parameters
    ----------
    infile : string
        The input template to use.
    outfile : string
        The file to create.
    timestep : float
        The time-step to use for the simulation.
    subcycles : integer
        The number of sub-cycles to perform.
    posfile : string
        The (base)name for the input file to read positions from.
    name : string, optional
        A name for the CP2K project.
    print_freq : integer, optional
        How often we should print to the trajectory file.

    """
    if print_freq is None:
        print_freq = subcycles
    to_update = {
        'GLOBAL': {
            'data': [f'PROJECT {name}',
                     'RUN_TYPE MD',
                     'PRINT_LEVEL LOW'],
            'replace': True,
        },
        'MOTION->MD':  {
            'data': {'STEPS': subcycles,
                     'TIMESTEP': timestep}
        },
        'MOTION->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
        'MOTION->PRINT->RESTART->EACH': {
            'data': {'MD': print_freq}
        },
        'MOTION->PRINT->VELOCITIES->EACH': {
            'data': {'MD': print_freq}
        },
        'MOTION->PRINT->TRAJECTORY->EACH': {
            'data': {'MD': print_freq}
        },
        'FORCE_EVAL->SUBSYS->TOPOLOGY': {
            'data': {'COORD_FILE_NAME': posfile,
                     'COORD_FILE_FORMAT': 'xyz'}
        },
        'FORCE_EVAL->DFT->SCF->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
    }
    remove = [
        'EXT_RESTART',
        'FORCE_EVAL->SUBSYS->COORD'
    ]
    update_cp2k_input(infile, outfile, update=to_update, remove=remove)


def write_for_continue(infile, outfile, timestep, subcycles,
                       name='md_continue'):
    """
    Create input file for a single step.

    Note, the single step actually consists of a number of subcycles.
    But from PyRETIS' point of view, this is a single step.
    Here, we make use of restart files named ``previous.restart``
    and ``previous.wfn`` to continue a run.

    Parameters
    ----------
    infile : string
        The input template to use.
    outfile : string
        The file to create.
    timestep : float
        The time-step to use for the simulation.
    subcycles : integer
        The number of sub-cycles to perform.
    name : string, optional
        A name for the CP2K project.

    """
    to_update = {
        'GLOBAL': {
            'data': [f'PROJECT {name}',
                     'RUN_TYPE MD',
                     'PRINT_LEVEL LOW'],
            'replace': True,
        },
        'MOTION->MD':  {
            'data': {'STEPS': subcycles,
                     'TIMESTEP': timestep}
        },
        'MOTION->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
        'MOTION->PRINT->RESTART->EACH': {
            'data': {'MD': subcycles}
        },
        'MOTION->PRINT->VELOCITIES->EACH': {
            'data': {'MD': subcycles}
        },
        'MOTION->PRINT->TRAJECTORY->EACH': {
            'data': {'MD': subcycles}
        },
        'EXT_RESTART': {
            'data': ['RESTART_VEL',
                     'RESTART_POS',
                     'RESTART_FILE_NAME previous.restart'],
            'replace': True
        },
        'FORCE_EVAL->DFT': {
            'data': {'WFN_RESTART_FILE_NAME': 'previous.wfn'},
        },
        'FORCE_EVAL->DFT->SCF->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
    }
    remove = [
        'FORCE_EVAL->SUBSYS->TOPOLOGY',
        'FORCE_EVAL->SUBSYS->VELOCITY',
        'FORCE_EVAL->SUBSYS->COORD'
        'FORCE_EVAL->DFT->RESTART_FILE_NAME',
    ]
    update_cp2k_input(infile, outfile, update=to_update, remove=remove)


def write_for_genvel(infile, outfile, posfile, seed,
                     name='genvel'):  # pragma: no cover
    """Create input file for velocity generation.

    2022.05.25: Note: This function is no longer in use
    as we now generate velocities internally. However we keep
    this function in case of future need.

    Parameters
    ----------
    infile : string
        The input template to use.
    outfile : string
        The file to create.
    posfile : string
        The (base)name for the input file to read positions from.
    seed : integer
        A seed for generating velocities.
    name : string, optional
        A name for the CP2K project.

    """
    to_update = {
        'GLOBAL': {
            'data': [f'PROJECT {name}',
                     f'SEED {seed}',
                     'RUN_TYPE MD',
                     'PRINT_LEVEL LOW'],
            'replace': True,
        },
        'FORCE_EVAL->DFT->SCF': {
            'data': {'SCF_GUESS': 'ATOMIC'}
        },
        'MOTION->MD':  {
            'data': {'STEPS': 1,
                     'TIMESTEP': 0}
        },
        'MOTION->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
        'MOTION->PRINT->RESTART->EACH': {
            'data': {'MD': 1}
        },
        'MOTION->PRINT->VELOCITIES->EACH': {
            'data': {'MD': 1}
        },
        'MOTION->PRINT->TRAJECTORY->EACH': {
            'data': {'MD': 1}
        },
        'FORCE_EVAL->SUBSYS->TOPOLOGY': {
            'data': {'COORD_FILE_NAME': posfile,
                     'COORD_FILE_FORMAT': 'xyz'}
        },
        'FORCE_EVAL->DFT->SCF->PRINT->RESTART': {
            'data': ['BACKUP_COPIES 0'],
            'replace': True,
        },
    }
    remove = [
        'EXT_RESTART',
        'FORCE_EVAL->SUBSYS->VELOCITY',
        'FORCE_EVAL->DFT->RESTART_FILE_NAME',
    ]
    update_cp2k_input(infile, outfile, update=to_update, remove=remove)


class CP2KEngine(ExternalMDEngine):
    """
    A class for interfacing CP2K.

    This class defines the interface to CP2K.

    Attributes
    ----------
    cp2k : string
        The command for executing CP2K.
    input_path : string
        The directory where the input files are stored.
    timestep : float
        The time step used in the CP2K MD simulation.
    subcycles : integer
        The number of steps each CP2K run is composed of.
    rgen : object like :py:class:`.RandomGenerator`
        An object we use to set seeds for velocity generation.
    extra_files : list
        List of extra files which may be required to run CP2K.

    """

    def __init__(self, cp2k, input_path, timestep, subcycles,
                 extra_files=None, exe_path=os.path.abspath('.'),  seed=0):
        """Set up the CP2K engine.

        Parameters
        ----------
        cp2k : string
            The CP2K executable.
        input_path : string
            The path to where the input files are stored.
        timestep : float
            The time step used in the CP2K simulation.
        subcycles : integer
            The number of steps each CP2K run is composed of.
        extra_files : list
            List of extra files which may be required to run CP2K.
        seed : integer, optional
            A seed for the random number generator.
        extra_files : list
            List of extra files which may be required to run CP2K.
        exe_path: string, optional
            The path on which the engine is executed

        """
        super().__init__('CP2K external engine', timestep,
                         subcycles)
        self.rgen = create_random_generator({'seed': seed})
        self.ext = 'xyz'
        self.cp2k = shlex.split(cp2k)
        logger.info('Command for execution of CP2K: %s', ' '.join(self.cp2k))
        # Store input path:
        self.input_path = os.path.join(exe_path, input_path)
        # Set the defaults input files:
        default_files = {
            'conf': f'initial.{self.ext}',
            'template': 'cp2k.inp',
        }
        # Check the presence of the defaults input files or, if absent,
        # try to find then by extension.
        self.input_files = look_for_input_files(self.input_path, default_files)

        # todo, these info can be processed by look_for_input_files using
        # the extra_files option.
        self.extra_files = []
        if extra_files is not None:
            for key in extra_files:
                fname = os.path.join(self.input_path, key)
                if not os.path.isfile(fname):
                    logger.critical('Extra CP2K input file "%s" not found!',
                                    fname)
                else:
                    self.extra_files.append(fname)

    def run_cp2k(self, input_file, proj_name):
        """
        Run the CP2K executable.

        Returns
        -------
        out : dict
            The files created by the run.

        """
        cmd = self.cp2k + ['-i', input_file]
        logger.debug('Executing CP2K %s: %s', proj_name, input_file)
        self.execute_command(cmd, cwd=self.exe_dir, inputs=None)
        out = {}
        for key, name in OUTPUT_FILES.items():
            out[key] = os.path.join(self.exe_dir, name.format(proj_name))
        return out

    def _extract_frame(self, traj_file, idx, out_file):
        """
        Extract a frame from a trajectory file.

        This method is used by `self.dump_config` when we are
        dumping from a trajectory file. It is not used if we are
        dumping from a single config file.

        Parameters
        ----------
        traj_file : string
            The trajectory file to dump from.
        idx : integer
            The frame number we look for.
        out_file : string
            The file to dump to.

        """
        for i, snapshot in enumerate(read_xyz_file(traj_file)):
            if i == idx:
                box, xyz, vel, names = convert_snapshot(snapshot)
                if os.path.isfile(out_file):
                    logger.debug('CP2K will overwrite %s', out_file)
                write_xyz_trajectory(out_file, xyz, vel, names, box,
                                     append=False)
                return
        logger.error('CP2K could not extract index %i from %s!',
                     idx, traj_file)

    def _propagate_from(self, name, path, ensemble, msg_file, reverse=False):
        """
        Propagate with CP2K from the current system configuration.

        Here, we assume that this method is called after the propagate()
        has been called in the parent. The parent is then responsible
        for reversing the velocities and also for setting the initial
        state of the system.

        Parameters
        ----------
        name : string
            A name to use for the trajectory we are generating.
        path : object like :py:class:`.PathBase`
            This is the path we use to fill in phase-space points.
        ensemble : dict
            It contains the simulations info:

            * `system` : object like :py:class:`.System`
              The system to act on.
            * `engine` : object like :py:class:`.EngineBase`
              This is the integrator that is used to propagate the system
              in time.
            * `order_function` : object like :py:class:`.OrderParameter`
              The class used for calculating the order parameters.
            * `interfaces` : list of floats
              These defines the interfaces for which we will check the
              crossing(s).

        msg_file : object like :py:class:`.FileIO`
            An object we use for writing out messages that are useful
            for inspecting the status of the current propagation.
        reverse : boolean, optional
            If True, the system will be propagated backward in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the
            propagation.

        """
        status = f'propagating with CP2K (reverse = {reverse})'
        system = ensemble['system']
        interfaces = ensemble['interfaces']
        logger.debug(status)
        success = False
        left, _, right = interfaces
        logger.debug('Adding input files for CP2K')
        # First, copy the required input files:
        self.add_input_files(self.exe_dir)
        # Get positions and velocities from the input file.
        initial_conf = system.particles.get_pos()[0]
        box, xyz, vel, atoms = self._read_configuration(initial_conf)
        if box is None:
            box, _ = read_cp2k_box(self.input_files['template'])
        # Add CP2K input for a single step:
        step_input = os.path.join(self.exe_dir, 'step.inp')
        write_for_step_vel(self.input_files['template'], step_input,
                           self.timestep, self.subcycles,
                           os.path.basename(initial_conf),
                           vel, name=name)
        # And create the input file for continuing:
        continue_input = os.path.join(self.exe_dir, 'continue.inp')
        write_for_continue(self.input_files['template'], continue_input,
                           self.timestep, self.subcycles, name=name)
        # Get the order parameter before the run:
        order = self.calculate_order(ensemble, xyz=xyz, vel=vel, box=box)
        traj_file = os.path.join(self.exe_dir, f'{name}.{self.ext}')
        # Create a message file with some info about this run:
        msg_file.write(
            f'# Initial order parameter: {" ".join([str(i) for i in order])}'
        )
        msg_file.write(f'# Trajectory file is: {traj_file}')
        # Run the first step:
        msg_file.write('# Running first CP2k step.')
        out_files = self.run_cp2k('step.inp', name)
        restart_file = os.path.join(self.exe_dir, out_files['restart'])
        prestart_file = os.path.join(self.exe_dir, 'previous.restart')
        wave_file = os.path.join(self.exe_dir, out_files['wfn'])
        pwave_file = os.path.join(self.exe_dir, 'previous.wfn')

        # Note: Order is calculated at the END of each iteration!
        i = 0
        # Write the config so we have a non-empty file:
        write_xyz_trajectory(traj_file, xyz, vel, atoms, box, step=i,
                             append=False)
        msg_file.write('# Running main CP2k propagation loop.')
        msg_file.write('# Step order parameter cv1 cv2 ...')
        for i in range(path.maxlen):
            msg_file.write(f'{i} {" ".join([str(j) for j in order])}')
            snapshot = {'order': order, 'config': (traj_file, i),
                        'vel_rev': reverse}
            phase_point = self.snapshot_to_system(system, snapshot)
            status, success, stop, add = self.add_to_path(path, phase_point,
                                                          left, right)
            if add and i > 0:
                # Write the previous configuration:
                write_xyz_trajectory(traj_file, xyz, vel, atoms, box,
                                     step=i)
            if stop:
                logger.debug('CP2K propagation ended at %i. Reason: %s',
                             i, status)
                break
            if i == 0:
                pass
            elif i > 0:
                self._movefile(restart_file, prestart_file)
                self._movefile(wave_file, pwave_file)
                if i < path.maxlen - 1:
                    out_files = self.run_cp2k('continue.inp', name)
            self._remove_files(self.exe_dir,
                               self._find_backup_files(self.exe_dir))
            # Read config after the step
            if i < path.maxlen - 1:
                atoms, xyz, vel, box, _ = read_cp2k_restart(restart_file)
                order = self.calculate_order(ensemble,
                                             xyz=xyz, vel=vel, box=box)
        msg_file.write('# Propagation done.')
        energy_file = out_files['energy']
        msg_file.write(f'# Reading energies from: {energy_file}')
        energy = read_cp2k_energy(energy_file)
        end = (i + 1) * self.subcycles
        ekin = energy.get('ekin', [])
        vpot = energy.get('vpot', [])
        path.update_energies(ekin[:end:self.subcycles],
                             vpot[:end:self.subcycles])
        for _, files in out_files.items():
            self._removefile(files)
        self._removefile(prestart_file)
        self._removefile(pwave_file)
        self._removefile(continue_input)
        self._removefile(step_input)
        return success, status

    def integrate(self, ensemble, steps, thermo='full'):
        """
        Propagate several integration steps.

        This method will perform several integration steps using
        CP2K. It will also calculate the order parameter(s) and
        energy terms if requested.

        Parameters
        ----------
        ensemble: dict
            it contains:

            * `system` : object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function` : object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.

        steps : integer
            The number of steps we are going to perform. Note that we
            do not integrate on the first step (e.g. step 0) but we do
            obtain the other properties. This is to output the starting
            configuration.
        thermo : string, optional
            Select the thermodynamic properties we are to calculate.

        Yields
        ------
        results : dict
            The results from a MD step. This contains the state of the system
            and order parameter(s) and energies (if calculated).

        """
        logger.debug('Integrating with CP2K')
        # First, copy the required input files:
        logger.debug('Adding input files for CP2K')
        self.add_input_files(self.exe_dir)
        system = ensemble['system']
        order_function = ensemble.get('order_function')
        # Get positions and velocities from the input file.
        initial_conf = self.dump_frame(system)

        box, xyz, vel, atoms = self._read_configuration(initial_conf)
        if box is None:
            box, _ = read_cp2k_box(self.input_files['template'])

        name = 'pyretis-cp2k'

        logger.debug('Thermo was set to: %s', thermo)

        # Add CP2K input for a single step:
        step_input = os.path.join(self.exe_dir, 'step.inp')
        write_for_integrate(self.input_files['template'], step_input,
                            self.timestep, self.subcycles,
                            os.path.basename(initial_conf),
                            name=name)
        # And create the input file for continuing:
        continue_input = os.path.join(self.exe_dir, 'continue.inp')
        write_for_continue(self.input_files['template'], continue_input,
                           self.timestep, self.subcycles, name=name)
        # Get the order parameter before the run:
        if order_function:
            order = self.calculate_order(ensemble,
                                         xyz=xyz, vel=vel, box=box)
        else:
            order = None
        traj_file = os.path.join(self.exe_dir, f'{name}.{self.ext}')
        # Run the first step:
        out_files = self.run_cp2k('step.inp', name)
        restart_file = os.path.join(self.exe_dir, out_files['restart'])
        prestart_file = os.path.join(self.exe_dir, 'previous.restart')
        wave_file = os.path.join(self.exe_dir, out_files['wfn'])
        pwave_file = os.path.join(self.exe_dir, 'previous.wfn')
        # Note: Order is calculated at the END of each iteration!
        i = 0
        # Write the config so we have a non-empty file:
        write_xyz_trajectory(traj_file, xyz, vel, atoms, box, step=i,
                             append=False)
        for i in range(steps):
            if i > 0:
                # Write the previous configuration:
                write_xyz_trajectory(traj_file, xyz, vel, atoms, box,
                                     step=i)
            if i == 0:
                pass
            elif i > 0:
                self._movefile(restart_file, prestart_file)
                self._movefile(wave_file, pwave_file)
                if i < steps - 1:  # Do not integrate the last step.
                    out_files = self.run_cp2k('continue.inp', name)
            self._remove_files(self.exe_dir,
                               self._find_backup_files(self.exe_dir))
            results = {}
            if order:
                results['order'] = order
            # Read config after the step
            if i < steps - 1:
                atoms, xyz, vel, box, _ = read_cp2k_restart(restart_file)
                if order_function:
                    order = self.calculate_order(
                        ensemble, xyz=xyz, vel=vel, box=box
                    )
            energy = read_cp2k_energy(out_files['energy'])
            end = (i + 1) * self.subcycles
            results['thermo'] = {}
            for key, val in energy.items():
                results['thermo'][key] = val[:end:self.subcycles][-1]
            yield results
        # Note we do not remove the run-files here as they might be
        # useful for the user.

    def step(self, system, name):
        """Perform a single step with CP2K.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we are integrating.
        name : string
            To name the output files from the CP2K step.

        Returns
        -------
        out : string
            The name of the output configuration, obtained after
            completing the step.

        """
        initial_conf = self.dump_frame(system)
        # Prepare input files etc.:
        self.add_input_files(self.exe_dir)
        box, xyz, vel, atoms = self._read_configuration(initial_conf)
        if box is None:
            box, _ = read_cp2k_box(self.input_files['template'])
        # Add CP2K input for a single step:
        step_input = os.path.join(self.exe_dir, 'step.inp')
        write_for_step_vel(self.input_files['template'], step_input,
                           self.timestep, self.subcycles,
                           os.path.basename(initial_conf),
                           vel, name=name)

        # Execute single step CP2K...
        logger.debug('Executing CP2K')
        out_files = self.run_cp2k('step.inp', name)
        energy = read_cp2k_energy(out_files['energy'])
        # Get the output configuration:
        atoms, xyz, vel, box, _ = read_cp2k_restart(out_files['restart'])
        conf_out = os.path.join(self.exe_dir, f'{name}.{self.ext}')
        write_xyz_trajectory(conf_out, xyz, vel, atoms, box, append=False)
        system.particles.set_pos((conf_out, None))
        system.particles.set_vel(False)
        system.particles.ekin = energy['ekin'][-1]
        system.particles.vpot = energy['vpot'][-1]
        system.update_box(box)
        # Prepare input files etc.:
        logger.debug('Removing CP2K output after single step.')
        # Remove run-files etc:
        for _, files in out_files.items():
            self._removefile(files)
        self._remove_files(
            self.exe_dir,
            self._find_backup_files(self.exe_dir)
        )
        return conf_out

    def add_input_files(self, dirname):
        """Add required input files to a given directory.

        Parameters
        ----------
        dirname : string
            The full path to where we want to add the files.

        """
        for files in self.extra_files:
            basename = os.path.basename(files)
            dest = os.path.join(dirname, basename)
            if not os.path.isfile(dest):
                logger.debug('Adding input file "%s" to "%s"',
                             basename, dirname)
                self._copyfile(files, dest)

    @staticmethod
    def _find_backup_files(dirname):
        """Return backup-files in the given directory."""
        out = []
        for entry in os.scandir(dirname):
            if entry.is_file():
                match = REGEXP_BACKUP.search(entry.name)
                if match is not None:
                    out.append(entry.name)
        return out

    @staticmethod
    def _read_configuration(filename):
        """
        Read CP2K output configuration.

        This method is used when we calculate the order parameter.

        Parameters
        ----------
        filename : string
            The file to read the configuration from.

        Returns
        -------
        box : numpy.array
            The box dimensions if we manage to read it.
        xyz : numpy.array
            The positions.
        vel : numpy.array
            The velocities.
        names : list of strings
            The atom names found in the file.

        """
        xyz, vel, box, names = None, None, None, None
        for snapshot in read_xyz_file(filename):
            box, xyz, vel, names = convert_snapshot(snapshot)
            break  # Stop after the first snapshot.
        return box, xyz, vel, names

    def _reverse_velocities(self, filename, outfile):
        """Reverse velocity in a given snapshot.

        Parameters
        ----------
        filename : string
            The configuration to reverse velocities in.
        outfile : string
            The output file for storing the configuration with
            reversed velocities.

        """
        box, xyz, vel, names = self._read_configuration(filename)
        write_xyz_trajectory(outfile, xyz, -1.0*vel, names, box, append=False)

    def _prepare_shooting_point(self, input_file):  # pragma: no cover
        """
        Create initial configuration for a shooting move.

        This creates a new initial configuration with random velocities.

        2022.05.25: Note: This function is no longer in use
        as we now generate velocities internally. However we keep
        this function in case of future need.

        Parameters
        ----------
        input_file : string
            The input configuration to generate velocities for.

        Returns
        -------
        output_file : string
            The name of the file created.
        energy : dict
            The energy terms read from the CP2K energy file.

        """
        box, xyz, vel, atoms = self._read_configuration(input_file)
        if box is None:
            box, _ = read_cp2k_box(self.input_files['template'])
        input_config = os.path.join(self.exe_dir, 'genvel_input.xyz')
        write_xyz_trajectory(input_config, xyz, vel, atoms, box, append=False)
        # Create input file for CP2K:
        run_file = os.path.join(self.exe_dir, 'run.inp')
        write_for_genvel(self.input_files['template'],
                         run_file,
                         'genvel_input.xyz',
                         self.rgen.random_integers(1, 999999999),
                         name='genvel')
        # Prepare to run it:
        self.add_input_files(self.exe_dir)
        out_files = self.run_cp2k(run_file, 'genvel')
        energy = read_cp2k_energy(out_files['energy'])
        # Get the output configuration:
        atoms, xyz, vel, box, _ = read_cp2k_restart(out_files['restart'])
        conf_out = os.path.join(self.exe_dir, f'genvel.{self.ext}')
        write_xyz_trajectory(conf_out, xyz, vel, atoms, box, append=False)
        # Remove run-files etc:
        for _, files in out_files.items():
            self._removefile(files)
        self._remove_files(
            self.exe_dir,
            self._find_backup_files(self.exe_dir)
        )
        return conf_out, energy

    def modify_velocities(self, ensemble, vel_settings=None):
        """
        Modify the velocities of the current state.

        This method will modify the velocities of a time slice.

        Parameters
        ----------
        ensemble: dict
            It contains:

            * `system` : object like :py:class:`.System`
              System is used here since we need access to the particle list.
            * `rgen` : object like :py:class:`.RandomGenerator`
              This is the random generator that will be used.

        vel_settings: dict, optional
            It contains info about the velocity settings:

            * `sigma_v` : numpy.array
              These values can be used to set a standard deviation (one
              for each particle) for the generated velocities.
            * `aimless` : boolean
               Determines if we should do aimless shooting or not.
            * `momentum` : boolean
              If True, we reset the linear momentum to zero after
              generating.
            * `rescale or rescale_energy` : float
              In some NVE simulations, we may wish to re-scale the
              energy to a fixed value. If `rescale` is a float > 0,
              we will re-scale the energy (after modification of
              the velocities) to match the given float.

        Returns
        -------
        dek : float
            The change in the kinetic energy.
        kin_new : float
            The new kinetic energy.

        """
        rgen = ensemble['rgen']
        system = ensemble['system']
        rescale = vel_settings.get('rescale_energy',
                                   vel_settings.get('rescale'))
        particles = system.particles
        pos = self.dump_frame(system)
        box, xyz, vel, atoms = self._read_configuration(pos)
        system.particles.vel = vel
        if box is None:
            box, _ = read_cp2k_box(self.input_files['template'])
        # to-do: retrieve particles.vpot from previous energy file.
        if None not in ((rescale, particles.vpot)) and rescale is not False:
            if rescale > 0:
                kin_old = rescale - particles.vpot
                do_rescale = True
            else:
                logger.warning('Ignored re-scale 6.2%f < 0.0.', rescale)
                return 0.0, calculate_kinetic_energy(particles)[0]
        else:
            kin_old = calculate_kinetic_energy(particles)[0]
            do_rescale = False
        if vel_settings.get('aimless', False):
            vel, _ = rgen.draw_maxwellian_velocities(system)
            system.particles.vel = vel
        else:
            dvel, _ = rgen.draw_maxwellian_velocities(
                system, sigma_v=vel_settings['sigma_v'])
            vel += dvel
            system.particles.vel = vel
        if vel_settings.get('momentum', False):
            reset_momentum(particles)
        if do_rescale:
            system.rescale_velocities(rescale, external=True)
        vel = system.particles.vel
        conf_out = os.path.join(self.exe_dir, f'genvel.{self.ext}')
        write_xyz_trajectory(conf_out, xyz, vel,
                             atoms, box, append=False)
        kin_new = calculate_kinetic_energy(system.particles)[0]
        system.particles.set_pos((conf_out, None))
        system.particles.set_vel(False)
        system.particles.ekin = kin_new
        if kin_old == 0.0:
            dek = float('inf')
            logger.debug(('Kinetic energy not found for previous point.'
                          '\n(This happens when the initial configuration '
                          'does not contain energies.)'))
        else:
            dek = kin_new - kin_old
        return dek, kin_new



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations.

Here we compare a full simulation with one where we have stopped
and restarted after 10 steps.
"""
import os
import sys
import colorama
import numpy as np
from pyretis.inout import print_to_screen
from pyretis.inout.settings import parse_settings_file
from pyretis.core.pathensemble import generate_ensemble_name
from pyretis.inout.formats.path import PathIntFile
from pyretis.inout.formats.energy import EnergyPathFile
from pyretis.inout.formats.order import OrderPathFile


def snapshot_difference(snap1, snap2):
    """Calculate difference between two snapshots."""
    diff = (snap1['pos'] - snap2['pos'])**2
    dsum = np.einsum('ij,ij -> i', diff, diff)
    diffv = (snap1['vel'] - snap2['vel'])**2
    dsumv = np.einsum('ij,ij -> i', diffv, diffv)
    return sum(dsum), sum(dsumv)


def compare_traj(traj1, traj2, tol=1e-12):
    """Compare two trajectories from PyRETIS.

    Here we calculate the mean squared error for the two
    trajectories.

    Parameters
    ----------
    traj1 : string
        A trajectory file to open.
    traj2 : string
        A trajectory file to open.
    tol : float
        A tolerance for comparing numbers.

    Returns
    -------
    out : int
        0 if comparison is successful, != 0 otherwise.

    """
    print_to_screen('Comparing trajectories', level='info')
    print_to_screen(f'Loading files: {traj1} & {traj2}')
    print_to_screen('Checking mean squared error...')
    file1 = PathIntFile(traj1, 'r').load()
    file2 = PathIntFile(traj2, 'r').load()
    error, error_v = [], []
    for trj1, trj2 in zip(file1, file2):
        for snap1, snap2 in zip(trj1['data'], trj2['data']):
            pose, vele = snapshot_difference(snap1, snap2)
            error.append(pose)
            error_v.append(vele)
    if next(file1, False) or next(file2, False):
        print_to_screen('Number of lines are incorrect ',
                        level='error')
        raise ValueError('Number of lines are incorrect')
    ret1 = print_error_assessment(np.mean(error), 'positions', tol)
    ret2 = print_error_assessment(np.mean(error_v), 'velocities', tol)
    return ret1 + ret2


def print_error_assessment(error, what, tol):
    """Print out some error info."""
    if abs(error) < tol:
        lev = 'success'
        ret = 0
    else:
        lev = 'error'
        ret = 1
    print_to_screen(f'Mean error - {what}: {error}',
                    level=lev)
    return ret


def compare_energy(traj1, traj2, tol=1e-12):
    """Compare energies from two PyRETIS trajectories.

    Here we calculate the mean squared error for the given files.

    Parameters
    ----------
    traj1 : string
        A trajectory file to open.
    traj2 : string
        A trajectory file to open.

    Returns
    -------
    out : int
        0 if comparison is successful, != 0 otherwise.

    """
    print_to_screen('Comparing energies', level='info')
    print_to_screen(f'Loading files: {traj1} & {traj2}')
    print_to_screen('Checking mean squared error...')
    file1 = EnergyPathFile(traj1, 'r').load()
    file2 = EnergyPathFile(traj2, 'r').load()
    errors = {}
    nsnap = 0
    for trj1, trj2 in zip(file1, file2):
        for key, values in trj1['data'].items():
            if key == 'time':
                continue
            if key not in errors:
                errors[key] = 0.0
            diff = (values - trj2['data'][key])**2
            errors[key] += sum(diff)
            nsnap += 1
    if next(file1, False) or next(file2, False):
        print_to_screen('Number of lines are incorrect ',
                        level='error')
        return 1
    retval = 0
    for key, err in errors.items():
        error = err / float(nsnap)
        ret = print_error_assessment(error, key, tol)
        retval += ret
    return retval


def compare_order(traj1, traj2, tol=1e-12):
    """Compare order parameters from two PyRETIS trajectories.

    Here we calculate the mean squared error for the given files.

    Parameters
    ----------
    traj1 : string
        A trajectory file to open.
    traj2 : string
        A trajectory file to open.

    Returns
    -------
    out : int
        0 if comparison is successful, != 0 otherwise.

    """
    print_to_screen('Comparing order parameters', level='info')
    print_to_screen(f'Loading files: {traj1} & {traj2}')
    print_to_screen('Checking mean squared error...')
    file1 = OrderPathFile(traj1, 'r').load()
    file2 = OrderPathFile(traj2, 'r').load()
    errors = {}
    nsnap = 0

    for trj1, trj2 in zip(file1, file2):
        col = tuple(trj1['data'].shape)
        for key in range(col[1]):
            if key == 0:
                continue
            if key not in errors:
                errors[key] = 0.0
            diff = (trj1['data'][:, key] - trj2['data'][:, key])**2
            errors[key] += sum(diff)
            nsnap += 1

    if next(file1, False) or next(file2, False):
        print_to_screen('Number of lines are incorrect ',
                        level='error')
        return 1

    retval = 0
    for key, err in errors.items():
        error = err / float(nsnap)
        ret = print_error_assessment(error, f'orderp-{key}', tol)
        retval += ret
    return retval


def compare_ensemble(ensemble):
    """Run the comparison for an ensemble."""
    print_to_screen(f'Comparing for "{ensemble}"', level='info')
    traj1 = os.path.join('run-10-20', ensemble, 'traj.txt')
    traj2 = os.path.join('run-full', ensemble, 'traj.txt')
    print_to_screen()
    ret1 = compare_traj(traj1, traj2, tol=1e-12)
    ener1 = os.path.join('run-10-20', ensemble, 'energy.txt')
    ener2 = os.path.join('run-full', ensemble, 'energy.txt')
    print_to_screen()
    ret2 = compare_energy(ener1, ener2, tol=1e-12)
    order1 = os.path.join('run-10-20', ensemble, 'order.txt')
    order2 = os.path.join('run-full', ensemble, 'order.txt')
    print_to_screen()
    ret3 = compare_order(order1, order2, tol=1e-12)
    return ret1 + ret2 + ret3


def main():
    """Run all comparisons."""
    settings = parse_settings_file(os.path.join('run-full', 'tis-001.rst'))
    ens = generate_ensemble_name(settings['tis'].get('ensemble_number'))
    ret = compare_ensemble(ens)
    if ret == 0:
        print_to_screen('\nAll seems fine!', level='success')
    else:
        print_to_screen('\nComparison failed!', level='error')
    return ret


if __name__ == '__main__':
    colorama.init(autoreset=True)
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of numerical integrators.

These integrators are typically used to integrate and propagate
Newtons equations of motion in time, the dynamics in molecular dynamics.
"""
import logging
from pyretis.engines import MDEngine
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class VVIntegrator(MDEngine):
    """VVIntegrator(MDEngine).

    This class defines the Velocity Verlet integrator.

    Attributes
    ----------
    timestep : float
        The time step.
    half_timestep : float
        Half of the time step.
    description : string
        Description of the integrator.

    """

    def __init__(self, timestep,
                 description='The velocity verlet integrator'):
        """Set up the Velocity Verlet integrator.

        Parameters
        ----------
        timestep : float
            The time step in internal units.
        description : string
            Description of the integrator.

        """
        super().__init__(timestep, description, dynamics='NVE')
        self.half_timestep = self.timestep * 0.5

    def integration_step(self, system):
        """Velocity Verlet integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in `system.particles`.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        imass = particles.imass
        particles.vel += self.half_timestep * particles.force * imass
        particles.pos += self.timestep * particles.vel
        system.potential_and_force()
        particles.vel += self.half_timestep * particles.force * imass


class Euler(MDEngine):
    """Euler(MDEngine).

    This class defines the Euler integrator.

    Attributes
    ----------
    timestep : float
        The time step.
    half_timestepsq : float
        Half of the squared time step.
    description : string
        Description of the integrator.

    """

    def __init__(self, timestep, description='The Euler integrator'):
        """Set up the Euler integrator.

        Parameters
        ----------
        timestep : float
            The time step in internal units.
        description : string
            Description of the integrator.

        """
        super().__init__(timestep, description, dynamics='NVE?')
        self.half_timestepsq = 0.5 * self.timestep**2

    def integration_step(self, system):
        """Euler integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in `system.particles`.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        imass = particles.imass
        # update positions and velocities
        particles.pos += (self.timestep * particles.vel +
                          self.half_timestepsq * particles.force * imass)
        particles.vel += self.timestep * particles.force * imass
        # update forces
        system.potential_and_force()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definitions of generic simulation objects.

This module defines the generic simulation object. This is the base
class for all other simulations.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Simulation (:py:class:`.Simulation`)
    A class defining a generic simulation.

"""
import logging
import copy
import os
from pyretis.simulation.simulation_task import SimulationTask
from pyretis.inout.screen import print_to_screen
from pyretis.inout.simulationio import task_from_settings
from pyretis.inout.restart import write_restart_file
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['Simulation']


class Simulation:
    """This class defines a generic simulation.

    Attributes
    ----------
    cycle : dict of integers
        This dictionary stores information about the number of cycles.
        The keywords are:

         *  `step`: The current cycle number.
         *  `startcycle`: The cycle number we started at.
         *  `endcycle`: Represents the cycle number where the simulation
            should end.
         * `stepno`: The number of cycles we have performed to arrive at
            cycle number given by `cycle['step']`. Note that `cycle['stepno']`
            might be different from `cycle['step']` since `cycle['startcycle']`
            might be != 0.

    exe_dir : string
        The path we are running the simulation from.
    restart_freq : integer
        The frequency for creating restart files.
    first_step : boolean
        True if the first step has not been executed yet.
    system : object like :py:class:`.System`
        This is the system the simulation will act on.
    simulation_output : list of dicts
        This list defines the output tasks associated
        with the simulation.
    simulation_type : string
        An identifier for the simulation.
    tasks : list of objects like :py:class:`.SimulationTask`
        This is the list of simulation tasks to execute.

    """

    simulation_type = 'generic'
    simulation_output = []

    def __init__(self, settings, controls):
        """Initialise the simulation object.

        Parameters
        ----------
        controls: dict of parameters to set up and control the simulations.
            It contains:

             *  `steps`: int, optional
                The number of simulation steps to perform.
             *  `startcycle`: int, optional
                The cycle we start the simulation on, useful for restarts.
             *  `endcycle`: int, optional
                The cycle we end the simulation to, useful in restarts.
             *  `rgen`: object like :py:class:`.RandomGenerator`
                The random generator that will be used for the
                paths that required random numbers.

        settings : dict
            Contains all the simulation settings.

        """
        steps = controls.get('steps', 0)
        startcycle = controls.get('startcycle', 0)
        end = controls.get('endcycle', steps)
        self.cycle = {'step': startcycle, 'endcycle': end,
                      'startcycle': startcycle, 'stepno': 0, 'steps': steps}

        self.tasks = []
        self.output_tasks = []
        self.first_step = True
        self.system = None
        self.restart_freq = None
        self.exe_dir = None
        self.settings = settings

    def extend_cycles(self, steps):
        """Extend a simulation with the given number of steps.

        Parameters
        ----------
        steps :  int
            The number of steps to extend the simulation with.

        Returns
        -------
        out : None
            Returns `None` but modifies `self.cycle`.

        """
        self.cycle['startcycle'] = self.cycle['stepno']
        self.cycle['endcycle'] = self.cycle['startcycle'] + steps

    def is_finished(self):
        """Determine if the simulation is finished.

        In this object, the simulation is done if the current step
        number is larger than the end cycle. Note that the number of
        steps performed is dependent on the value of
        `self.cycle['startcycle']`.

        Returns
        -------
        out : boolean
            True if the simulation is finished, False otherwise.

        """
        return self.cycle['step'] >= self.cycle['endcycle']

    def step(self):
        """Execute a simulation step.

        Here, the tasks in :py:attr:`.tasks` will be executed
        sequentially.

        Returns
        -------
        out : dict
            This dictionary contains the results of the defined tasks.

        Note
        ----
        This function will have 'side effects' and update/change
        the state of other attached variables such as the system or
        other variables that are not explicitly shown. This is intended
        and the behavior is defined by the tasks in
        :py:attr:`.tasks`.

        """
        if not self.first_step:
            self.cycle['step'] += 1
            self.cycle['stepno'] += 1
        results = self.execute_tasks()
        if self.first_step:
            self.first_step = False
        return results

    def execute_tasks(self):
        """Execute all the tasks in sequential order.

        Returns
        -------
        results : dict
            The results from the different tasks (if any).

        """
        results = {'cycle': self.cycle.copy()}
        for task in self.tasks:
            if not self.first_step or task.run_first():
                resi = task.execute(self.cycle)
                if task.result is not None:
                    results[task.result] = resi
        results['system'] = self.system
        return results

    def add_task(self, task, position=None):
        """Add a new simulation task.

        A task can still be added manually by simply appending to
        py:attr:`.tasks`. This function will, however, do some
        checks so that the task added can be executed.

        Parameters
        ----------
        task : dict
            A dict defining the task. A task is represented by an
            object of type :py:class:`.SimulationTask` with some
            additional settings on how to store the output
            and when to execute the task.
            The keywords in the dict defining the task are:

            * `func`: Callable, this is a function to execute in the
              task.
            * `args`: List, with arguments for the function.
            * `kwargs`: Dict, with the keyword arguments for the
              function.
            * `when`: Dict, which defines when the task should be
              executed.
            * `first`: Boolean, determines if the task should be
              executed on the initial step, i.e. before the
              full simulation starts.
            * `result`: String, used to label the result.

        position : int, optional
            Can be used to give the tasks a specific position in the
            task list.

        """
        try:
            new_task = SimulationTask(task['func'],
                                      args=task.get('args', None),
                                      kwargs=task.get('kwargs', None),
                                      when=task.get('when', None),
                                      result=task.get('result', None),
                                      first=task.get('first', False))
            if position is None:
                self.tasks.append(new_task)
            else:
                self.tasks.insert(position, new_task)
            return True
        except AssertionError:
            logger.warning('Could not add task: %s', task)
            return False

    def run(self):
        """Run a simulation.

        The intended usage is for simulations where all tasks have
        been defined in :py:attr:`self.tasks`.

        Note
        ----
        This function will just run the tasks via executing
        :py:meth:`.step` In general, this is probably too generic for
        the simulation you want, if you are creating a custom simulation.
        Please consider customizing the :py:meth:`.run` (or the
        :py:meth:`.step`) method of your simulation class.

        Yields
        ------
        out : dict
            This dictionary contains the results from the simulation.

        """
        while not self.is_finished():
            result = self.step()
            for task in self.output_tasks:
                task.output(result)
            self.write_restart()
            if self.soft_exit():
                yield result
                break
            yield result

    def __str__(self):
        """Just a small function to return some info about the simulation."""
        ntask = len(self.tasks)
        mtask = 'task' if ntask == 1 else 'tasks'
        msg = [f'Generic simulation with {ntask} {mtask}.']
        for i, task in enumerate(self.tasks):
            msg += [f'* Task no. {i}']
            for j, line in enumerate(str(task).split('\n')):
                if j > 0:
                    msg += [line]
        return '\n'.join(msg)

    def set_up_output(self, settings, progress=False):
        """Set up output from the simulation.

        This includes the predefined output tasks, but also output
        related to the restart file(s).


        Parameters
        ----------
        settings : dict
            These are the simulation settings.
        progress : boolean
            For some simulations, the user may select to display a
            progress bar, we then need to disable the screen output.

        """
        logging.debug('Setting up output for simulation %s',
                      self.__class__.__name__)
        # Create the output tasks:
        self.create_output_tasks(settings, progress=progress)
        # Do set-up for restart output:
        self.restart_freq = settings['output'].get('restart-file', -1)
        if self.restart_freq < 1:
            self.restart_freq = None
            logger.warning('Writing of restart file(s) disabled!')
        logger.debug('Setting restart frequency for simulation %s',
                     self.restart_freq)
        self.exe_dir = settings['simulation'].get('exe_path', os.getcwd())

    def create_output_tasks(self, settings, progress=False):
        """Create output tasks for the simulation.

        This method will generate output tasks based on the tasks
        listed in :py:attr:`.simulation_output`.

        Parameters
        ----------
        settings : dict
            These are the simulation settings.
        progress : boolean
            For some simulations, the user may select to display a
            progress bar, we then need to disable the screen output.

        """
        logging.debug('Clearing output tasks & adding pre-defined ones')
        engine = getattr(self, 'engine', None)
        order_function = getattr(self, 'order_function', None)
        self.output_tasks = []
        directory = settings['simulation'].get('exe_path', None)
        for task_dict in self.simulation_output:
            if 'order' in task_dict['type'] and order_function is None:
                continue
            task = task_from_settings(task_dict, settings, directory, engine,
                                      progress)
            if task is not None:
                logger.debug('Created output task:\n%s', task)
                self.output_tasks.append(task)

    def soft_exit(self):
        """Force simulation to stop at the current step."""
        exit_file = 'EXIT'
        if self.exe_dir:
            exit_file = os.path.join(self.exe_dir, exit_file)
        if os.path.isfile(exit_file):
            logger.info('Exit file found - will do a soft exit.')
            print_to_screen('Exit file found - will do a soft exit.',
                            level='warning')
            # Write restart file...
            self.write_restart(now=True)
            # Close output files...
            for task in self.output_tasks:
                if task.target == 'file':
                    task.writer.close()
            return True
        return False

    def write_restart(self, now=False):
        """Create a restart file.

        Parameters
        ----------
        now : boolean, optional
            If True, the output file will be written irrespective of the
            step number.

        """
        if now or (self.restart_freq is not None and
                   self.cycle['step'] % self.restart_freq == 0):
            out = 'pyretis.restart'
            if self.exe_dir:
                out = os.path.join(self.exe_dir, out)
            write_restart_file(out, self)

    def restart_info(self):
        """Return information which can be used to restart the simulation.

        Returns
        -------
        info : dict,
            Contains all the updated simulation settings and counters.

        """
        info = {}
        info['settings'] = self.settings

        for key in ('simulation', 'system', 'particles'):
            if key not in info:
                info[key] = {}

        info['simulation']['restart'] = 'pyretis.restart'
        info['simulation']['cycle'] = copy.deepcopy(self.cycle)
        info['simulation']['type'] = self.simulation_type

        if hasattr(self, 'system') and self.system is not None:
            info['system'].update(self.system.restart_info())

        if hasattr(self.system, 'particles') and \
                self.system.particles is not None:
            info['particles'].update(self.system.particles.restart_info())

        return info

    def load_restart_info(self, info):
        """Load restart information.

        Note, we do not change the ``end`` property here as we probably
        are extending a simulation.

        Parameters
        ----------
        info : dict
            The dictionary with the restart information.

        """
        for key, val in info['simulation']['cycle'].items():
            if key in {'steps', 'endcycle'}:
                self.cycle['startcycle'] = copy.deepcopy(val)
            else:
                self.cycle[key] = copy.deepcopy(val)

        if info.get('system') is not None and\
                hasattr(self, 'system') and self.system is not None:
            self.system.load_restart_info(info['system'])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of a dihedral order parameter.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Dihedral (:py:class:`.Dihedral`)
    A dihedral angle order parameter.
"""
import logging
from numpy import dot, cross, arctan2
from numpy.linalg import norm
from pyretis.orderparameter.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['Dihedral']


class Dihedral(OrderParameter):
    """Calculates the dihedral angle defined by 4 atoms.

    The angle definition is given by Blondel and Karplus,
    J. Comput. Chem., vol. 17, 1996, pp. 1132--1141. If we
    label the 4 atoms A, B, C and D, then the angle is given by
    the vectors u = A - B, v = B - C, w = D - C

    Attributes
    ----------
    index : list/tuple of integers
        These are the indices for the atoms to use in the
        definition of the dihedral angle.
    periodic : boolean
        This determines if periodic boundaries should be applied to
        the position or not.

    """

    def __init__(self, index, periodic=False):
        """Initialise the order parameter.

        Parameters
        ----------
        index : list/tuple of integers
            This list gives the indices for the atoms to use in the
            definition of the dihedral angle.
        periodic : boolean, optional
            This determines if periodic boundary conditions should be
            applied to the distance vectors.

        """
        try:
            if len(index) != 4:
                msg = ('Wrong number of atoms for dihedral definition. '
                       f'Expected 4 got {len(index)}')
                logger.error(msg)
                raise ValueError(msg)
        except TypeError as err:
            msg = 'Dihedral should be defined as a tuple/list of integers!'
            logger.error(msg)
            raise TypeError(msg) from err
        self.index = [int(i) for i in index]
        txt = ('Dihedral angle between particles '
               f'{index[0]}, {index[1]}, {index[2]} and {index[3]}')
        super().__init__(description=txt)
        self.periodic = periodic

    def calculate(self, system):
        """Calculate the dihedral angle.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The object containing the information we need to calculate
            the order parameter.

        Returns
        -------
        out : list of float
            The order parameter.

        """
        pos = system.particles.pos
        vector1 = pos[self.index[0]] - pos[self.index[1]]
        vector2 = pos[self.index[1]] - pos[self.index[2]]
        vector3 = pos[self.index[3]] - pos[self.index[2]]
        if self.periodic:
            vector1 = system.box.pbc_dist_coordinate(vector1)
            vector2 = system.box.pbc_dist_coordinate(vector2)
            vector3 = system.box.pbc_dist_coordinate(vector3)
        # Norm to simplify formulas:
        vector2 /= norm(vector2)
        denom = (dot(vector1, vector3) -
                 dot(vector1, vector2) * dot(vector2, vector3))
        numer = dot(cross(vector1, vector2), vector3)
        angle = arctan2(numer, denom)
        return [angle]



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the particle functions from pyretis.core.particlefunctions"""
import logging
import unittest
import numpy as np
from pyretis.core.box import create_box
from pyretis.core.system import System
from pyretis.core.particles import Particles
from pyretis.core.particlefunctions import (
    _get_vel_mass,
    atomic_kinetic_energy_tensor,
    calculate_kinetic_energy,
    calculate_kinetic_energy_tensor,
    calculate_kinetic_temperature,
    kinetic_temperature,
    calculate_linear_momentum,
    calculate_pressure_from_temp,
    calculate_pressure_tensor,
    calculate_scalar_pressure,
    calculate_thermo,
    calculate_thermo_path,
    reset_momentum,
)
logging.disable(logging.CRITICAL)


def empty_particles(masses):
    """Set up a particle object for testing."""
    particles = Particles(dim=3)
    for i in masses:
        particles.add_particle(np.zeros(3), np.zeros(3), np.zeros(3), mass=i)
    particles.virial = np.zeros((3, 3))
    return particles


class ParticleTest(unittest.TestCase):
    """Run the tests for the Particle() class."""

    def test_get_selection(self):
        """Test the creation a particle list."""
        particles = empty_particles([1, 2, 1, 2])
        _, mass = _get_vel_mass(particles, selection=[0, 1])
        for i, j in zip(mass, (1, 2)):
            self.assertAlmostEqual(i[0], j)
        _, mass = _get_vel_mass(particles, selection=[0, 2])
        for i in mass:
            self.assertAlmostEqual(i[0], 1.0)
        _, mass = _get_vel_mass(particles, selection=[1, 3])
        for i in mass:
            self.assertAlmostEqual(i[0], 2.0)

    def test_get_atomic_ek(self):
        """Test the atomic ek tensor calculation."""
        particles = empty_particles([1, 2, 1, 2])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.],
                                  [3., 3., 3.], [4., 4., 4.]])
        kin = atomic_kinetic_energy_tensor(particles)
        kin1 = atomic_kinetic_energy_tensor(particles, selection=[0, 2])
        kin2 = atomic_kinetic_energy_tensor(particles, selection=[1, 3])
        self.assertTrue(np.allclose(kin[0], kin1[0]))
        self.assertTrue(np.allclose(kin[2], kin1[1]))
        self.assertTrue(np.allclose(kin[1], kin2[0]))
        self.assertTrue(np.allclose(kin[3], kin2[1]))
        kin3 = atomic_kinetic_energy_tensor(particles, selection=[3])
        self.assertTrue(np.allclose(kin[3], kin3[0]))

    def test_calculate_ek(self):
        """Test calculation of kinetic energy."""
        particles = empty_particles([1, 2, 1, 2])
        particles.vel = np.array([[1., 1., 1.], [2.1, 2.1, 2.1],
                                  [3.1, 3.1, 3.1], [4.1, 4.1, 4.1]])
        ekin, tensor = calculate_kinetic_energy(particles)
        tensor2 = calculate_kinetic_energy_tensor(particles)
        self.assertTrue(np.allclose(tensor, tensor2))
        self.assertAlmostEqual(ekin, 79.575)

    def test_calculate_temp(self):
        """Test calculation of temperature."""
        particles = empty_particles([1, 2, 1, 0.01])
        particles.vel = np.ones_like(particles.pos)
        temp1, temp2, _ = calculate_kinetic_temperature(particles, 1.0)
        self.assertAlmostEqual(temp2, 1.0025)
        for i in temp1:
            self.assertAlmostEqual(i, 1.0025)

    def test_kinetic_temp(self):
        """Test calculation of kinetic temperature."""
        particles = empty_particles([1, 2, 1, 0.01])
        particles.vel = np.ones_like(particles.pos)
        vel = particles.vel
        mass = particles.mass
        temp1, temp2, _ = kinetic_temperature(vel, mass, 1.0)
        self.assertAlmostEqual(temp2, 1.0025)
        for i in temp1:
            self.assertAlmostEqual(i, 1.0025)
        temp1, temp2, _ = kinetic_temperature(vel, mass, 1.0,
                                              dof=[1.0, 0.0, 0.0])
        self.assertAlmostEqual(temp2, 1.11388888889)
        for i, j in zip(temp1, (1.33666667, 1.0025, 1.0025)):
            self.assertAlmostEqual(i, j)

    def test_linear_momentum(self):
        """Test calculation of linear momentum."""
        particles = empty_particles([1, 0.5, 2, 1])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.],
                                  [0.5, 0.5, 0.5], [1., 1., 1.]])
        mom = calculate_linear_momentum(particles)
        for i in mom:
            self.assertAlmostEqual(i, 4.)

    def test_pressure_from_temp(self):
        """Test calculation of pressure."""
        particles = empty_particles([1, 0.5, 2, 1])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.],
                                  [0.5, 0.5, 0.5], [1., 1., 1.]])
        pvol, press = calculate_pressure_from_temp(particles, 3, 1.0, 1.0)
        self.assertAlmostEqual(pvol, 4.5)
        self.assertAlmostEqual(press, 4.5)
        pvol, press = calculate_pressure_from_temp(particles, 3, 1.0, 1.0,
                                                   dof=[1., 0., 0.])
        self.assertAlmostEqual(pvol, 4.583333333)
        self.assertAlmostEqual(press, 4.583333333)

    def test_pressure_tensors(self):
        """Test calculation of the pressure tensor and scalar."""
        particles = empty_particles([1, 2, 3, 4])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.],
                                  [0.5, 0.5, 0.5], [1., 1., 1.]])
        press = calculate_pressure_tensor(particles, 1.0)
        for i in press.ravel():
            self.assertAlmostEqual(i, 13.75)
        presss = calculate_scalar_pressure(particles, 1.0, 3.0)
        self.assertAlmostEqual(presss, 13.75)

    def test_calculate_thermo(self):
        """Test the calculate_thermo method."""
        system = System(units='reduced', box=create_box(cell=[1., 1., 1.]))
        particles = empty_particles([1, 2, 3, 4])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.],
                                  [0.5, 0.5, 0.5], [1., 1., 1.]])
        particles.vpot = 123.456
        system.particles = particles
        res = calculate_thermo(system)
        correct = {
            'etot': 36.020250000000004,
            'vpot': 30.864,
            'press': 13.75,
            'mom': np.array([10.5, 10.5, 10.5]),
            'temp': 4.583333333333333,
            'ekin': 5.15625,
            'press-tens': 13.75 * np.ones((3, 3)),
        }
        for key in ('etot', 'vpot', 'press', 'temp', 'ekin'):
            self.assertAlmostEqual(correct[key], res[key])
        for key in ('mom', 'press-tens'):
            self.assertTrue(np.allclose(correct[key], res[key]))

    def test_thermo_path(self):
        """Test thermo calculation for path-style output."""
        system = System(units='reduced', box=create_box(cell=[1., 1.5, 1.]))
        particles = empty_particles([1, 2])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.]])
        particles.vpot = 314.21
        system.particles = particles
        res = calculate_thermo_path(system)
        correct = {
            'temp': 9.0,
            'vpot': 314.21,
            'ekin': 13.5,
            'etot': 327.70999999999998
        }
        for key, val in correct.items():
            self.assertAlmostEqual(val, res[key])

    def test_reset_mom(self):
        """Test that we can reset momentum."""
        particles = empty_particles([1, 0.5, 2, 1])
        particles.vel = np.array([[1., 1., 1.], [2., 2., 2.],
                                  [0.5, 0.5, 0.5], [1., 1., 1.]])
        vel = np.copy(particles.vel)
        reset_momentum(particles)
        mom = calculate_linear_momentum(particles)
        self.assertTrue(np.allclose(mom, np.zeros(3)))
        particles.vel = vel
        reset_momentum(particles, dim=[False, True, False])
        mom = calculate_linear_momentum(particles)
        correct = [4., 0.0, 4.0]
        for i, j in zip(mom, correct):
            self.assertAlmostEqual(i, j)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test of the WCA pair potential."""
import logging
import unittest
import numpy as np
from pyretis.core.box import create_box
from pyretis.core.particles import Particles
from pyretis.core.system import System
from pyretis.forcefield.potentials.pairpotentials.wca import (
    DoubleWellWCA,
)
logging.disable(logging.CRITICAL)


CORRECT_FORCE = np.array([-2416.67712532,
                          -2416.67712532,
                          -2416.67712532])
CORRECT_VIRIAL = np.ones((3, 3)) * 483.33542506
CORRECT_VPOT = 873.18173451


def create_system_box(in_max=True):
    """Create a simple test system."""
    box = create_box(cell=[10, 10, 10])
    system = System(box=box)
    system.particles = Particles(system.get_dim())
    system.add_particle(name='Ar1', pos=np.array([1.0, 1.0, 1.0]),
                        mass=1.0, ptype=0)
    if in_max:
        system.add_particle(name='Ar2', pos=np.array([2.25, 1.0, 1.0]),
                            mass=1.0, ptype=0)
    else:
        system.add_particle(name='Ar2', pos=np.array([1.2, 1.2, 1.2]),
                            mass=1.0, ptype=0)
    system.add_particle(name='X', pos=np.array([1.0, 2.0, 1.0]),
                        mass=1.0, ptype=1)
    system.add_particle(name='X', pos=np.array([1.0, 3.25, 1.0]),
                        mass=1.0, ptype=1)
    pot = DoubleWellWCA()
    parameters = {
        'types': {(0, 0)},
        'rzero': 1.0,
        'width': 0.25,
        'height': 6.0,
    }
    pot.set_parameters(parameters)
    return system, box, pot


class TestWCAPairPotential(unittest.TestCase):
    """Run tests for the WCA pair potential."""

    def test_parameters(self):
        """Test setup with parameters."""
        pot = DoubleWellWCA()
        parameters = {
            'types': {(0, 0)},
            'rzero': 1.0,
            'width': 0.25,
            'height': 6.0,
        }
        pot.set_parameters(parameters)
        for key, val in parameters.items():
            self.assertEqual(val, pot.params[key])
        # Check derived parameters:
        self.assertAlmostEqual(parameters['width']**2, pot.params['width2'])
        self.assertAlmostEqual(parameters['rzero'] + parameters['width'],
                               pot.params['rwidth'])
        self.assertAlmostEqual(parameters['height'] * 4.0,
                               pot.params['height4'])
        # Check that extra parameters are ignored:
        parameters['junk'] = 'should-not-be-added!'
        logging.disable(logging.INFO)
        mod = 'pyretis.forcefield.potentials.pairpotentials.wca'
        with self.assertLogs(mod, level='WARNING'):
            pot.set_parameters(parameters)
        logging.disable(logging.CRITICAL)
        self.assertFalse('junk' in pot.params)

    def test_activate(self):
        """Test that the WCA potential is activated for correct pairs."""
        pot = DoubleWellWCA()
        parameters = {
            'types': {(0, 0)},
            'rzero': 1.0,
            'width': 0.25,
            'height': 6.0,
        }
        pot.set_parameters(parameters)
        # pylint: disable=protected-access
        self.assertTrue(pot._activate(0, 0))
        self.assertFalse(pot._activate(0, 1))
        parameters['types'] = {(0, 0), (0, 1)}
        pot.set_parameters(parameters)
        self.assertTrue(pot._activate(0, 0))
        self.assertTrue(pot._activate(1, 0))
        self.assertTrue(pot._activate(0, 1))
        self.assertFalse(pot._activate(1, 1))
        parameters['types'] = None
        pot.set_parameters(parameters)
        self.assertTrue(pot._activate(0, 0))
        self.assertTrue(pot._activate(1, 0))
        self.assertTrue(pot._activate(0, 1))
        self.assertTrue(pot._activate(1, 1))

    def test_min_max(self):
        """Test that the min/max is correctly set for the WCA potential."""
        pot = DoubleWellWCA()
        parameters = {
            'types': {(0, 0)},
            'rzero': 1.0,
            'width': 0.25,
            'height': 6.0,
        }
        pot.set_parameters(parameters)
        min1, min2, max1 = pot.min_max()
        self.assertAlmostEqual(min1, 1.0)
        self.assertAlmostEqual(min2, 1.5)
        self.assertAlmostEqual(max1, 1.25)

    def test_potential(self):
        """Test that the potential is evaluated correctly."""
        system, _, pot = create_system_box()
        vpot = pot.potential(system)
        self.assertAlmostEqual(vpot, 6.0)

    def test_force(self):
        """Test that the potential and virial is correctly evaluated."""
        system, _, pot = create_system_box(in_max=False)
        force, virial = pot.force(system)
        self.assertTrue(np.allclose(force[0], CORRECT_FORCE))
        self.assertTrue(np.allclose(force[1], -1.0 * CORRECT_FORCE))
        self.assertTrue(np.allclose(force[2], np.zeros(3)))
        self.assertTrue(np.allclose(force[3], np.zeros(3)))
        self.assertTrue(np.allclose(virial, CORRECT_VIRIAL))

    def test_potential_and_force(self):
        """Test that the potential and virial is correctly evaluated."""
        system, _, pot = create_system_box(in_max=False)
        vpot, force, virial = pot.potential_and_force(system)
        self.assertTrue(np.allclose(force[0], CORRECT_FORCE))
        self.assertTrue(np.allclose(force[1], -1.0 * CORRECT_FORCE))
        self.assertTrue(np.allclose(force[2], np.zeros(3)))
        self.assertTrue(np.allclose(force[3], np.zeros(3)))
        self.assertTrue(np.allclose(virial, CORRECT_VIRIAL))
        self.assertAlmostEqual(vpot, CORRECT_VPOT)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the FileIO class."""
import logging
import numpy as np
import os
import unittest
from pyretis.analysis import match_all_histograms
from pyretis.inout.plotting import create_plotter
from pyretis.inout.report.report import generate_report, get_template
from .help import turn_on_logging

logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))


class Reporting(unittest.TestCase):
    """Test the generate_report."""
    def test_generate_report(self):
        """Test analysisio function"""
        with self.assertRaises(ValueError) as ext:
            generate_report(report_type='fake', analysis_results={},
                            output='ext')
        self.assertIn('nkown report type fake', str(ext.exception))
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.report.report',
                                 level='WARNING'):
                with self.assertRaises(ValueError) as ext:
                    generate_report(report_type='retis0',
                                    analysis_results={},
                                    output='ext')
                self.assertIn('Could not locate template', str(ext.exception))

    def test_create_plotter(self):
        """Test create_plotter."""
        with self.assertRaises(ValueError) as ext:
            create_plotter({'plotter': 'Fake'})
        self.assertIn('Unknown plotter: Fake', str(ext.exception))

        with self.assertRaises(ValueError) as ext:
            create_plotter({'plotter': 'mpl', 'output': 'kik'})
        self.assertIn('Output format "kik" is not support', str(ext.exception))

    def test_get_template(self):
        """Test get_template."""
        # Any file would do
        ifile = os.path.join(HERE, 'initial-gro.rst')
        nav, path = get_template(output='', report_type='', template=ifile)

        self.assertEqual((path, nav), (HERE, 'initial-gro.rst'))

    def test_match_all_histograms(self):
        """Test match histograms for umbrella sampling."""
        histograms = [[np.array([1, 5, 1, 0, 1, 1]),
                       np.array([+0.76, -0.74, -0.72, -0.71, +0.68, -0.66]),
                       np.array([-0.26, -0.34, +0.42, -0.52, -0.18, -0.16])],
                      [np.array([1, 3, 1, 0, 3, 1]),
                       np.array([-0.76, -0.74, +0.72, -0.71, -0.68, -0.66]),
                       np.array([-0.26, +0.34, -0.42, +0.32, +0.18, -0.16])],
                      [np.array([1, 2, 1, 0, 7, 1]),
                       np.array([-0.76, -0.74, +0.72, -0.71, -0.68, +0.66]),
                       np.array([-0.26, +0.34, -0.42, -0.22, -0.18, -0.16])]]
        windows = [[-1.0, -0.4], [-0.5, -0.2], [-0.3, 0.0]]

        histograms_s, mix, hist_avg = match_all_histograms(histograms, windows)
        self.assertTrue(all(histograms_s[0] == [1, 5, 1, 0, 1, 1]))
        self.assertTrue(all(histograms_s[1] == [1., 3., 1., 0., 3., 1.]))
        self.assertTrue(all(histograms_s[2] == [1., 2., 1., 0., 7., 1.]))
        self.assertTrue(all(np.array(mix) == [1.0, 1.0, 1.0]))
        self.assertTrue(all(hist_avg == np.array([1., 3., 0., 0., 7., 1.])))


if __name__ == '__main__':
    unittest.main()
